/*  
 * Copyright (c) 2009 Ant Kutschera, maxant
 * 
 * This file is part of the maxant GWT Demo.
 * 
 * maxant GWT Demo is free software: you can redistribute it and/or modify
 * it under the terms of the Lesser GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * maxant GWT Demo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Lesser GNU General Public License for more details.
 * You should have received a copy of the Lesser GNU General Public License
 * along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */

package uk.co.maxant.gwtdemo.client;

import java.util.List;

import com.google.gwt.user.client.rpc.RemoteService;
import com.google.gwt.user.client.rpc.RemoteServiceRelativePath;

import uk.co.maxant.gwtdemo.client.to.Artist;
import uk.co.maxant.gwtdemo.client.to.BasketItem;
import uk.co.maxant.gwtdemo.client.to.Design;
import uk.co.maxant.gwtdemo.client.to.DesignComment;
import uk.co.maxant.gwtdemo.client.to.OrderState;
import uk.co.maxant.gwtdemo.client.to.PaypalConfig;
import uk.co.maxant.gwtdemo.client.to.SalesOrder;
import uk.co.maxant.gwtdemo.client.to.SearchResults;
import uk.co.maxant.gwtdemo.client.to.SessionInfo;
import uk.co.maxant.gwtdemo.client.to.User;
import uk.co.maxant.gwtdemo.server.DesignService;
import uk.co.maxant.gwtdemo.server.OrchestrationService;
import uk.co.maxant.gwtdemo.server.SalesOrderService;
import uk.co.maxant.gwtdemo.server.PartyService;


/**
 */

@RemoteServiceRelativePath("services")
public interface ApplicationLayerService extends RemoteService {

    /**
     * gets a list of featured designs for the current domain (taken out of the request).
     *
     * @return  list of all designs
     *
     * @throws  ApplicationException  if anything goes wrong
     */
    List<Design> getFeaturedDesigns() throws ApplicationException;

    /**
     * gets a list of newest designs for the current domain (taken out of the request).
     *
     * @return  list of all designs
     *
     * @throws  ApplicationException  if anything goes wrong
     */
    List<Design> getNewestDesigns() throws ApplicationException;
    
    /**
     * adds the design to the basket.
     *
     * @param   basketItem  basket item comprising the design to add to the basket. may be null, in
     *                      which case this method simply returns the existing contents of the basket.
     *
     * @return  the contents of the basket
     *
     * @throws  ApplicationException  if anything goes wrong
     */
    List<BasketItem> addToBasket(BasketItem basketItem) throws ApplicationException;

    /**
     * gets all designs that are in the basket.
     *
     * @return  the contents of the basket
     *
     * @throws  ApplicationException  if anything goes wrong
     */
    List<BasketItem> getBasket() throws ApplicationException;

    /**
     * empties the basket.
     *
     * @return  the contents of the basket
     *
     * @throws  ApplicationException  if anything goes wrong
     */
    List<BasketItem> emptyBasket() throws ApplicationException;

    /**
     * creates the order persistently.
     *
     * @param   address  the address where to send the order to
     *
     * @return  the order ID
     *
     * @throws  ApplicationException  if anything goes wrong
     */
    Integer createOrder(String email, String name) throws ApplicationException;

    /**
     * gets session info including who is logged in.
     *
     * @return  info. login and other user info is null, if not logged in.
     *
     * @throws  ApplicationException  if anything goes wrong
     */
    SessionInfo getSessionInfo() throws ApplicationException;

    /**
     * gets the paypal config for the given order, so the browser can be redirected to paypal.
     *
     * @param   soUid  the order id that should be paid for
     *
     * @return  the config to be used to redirect to paypal
     *
     * @throws  ApplicationException  if anything goes wrong
     */
    PaypalConfig getPaypalConfig(Integer soUid) throws ApplicationException;

    /**
     * gets list of fully loaded designs only if they are in the catalogue.
     * 
     * @param	getAllContent normally false, but if true it then gets you non-purchaseables too. 
     * 					eg when an artist is viewing their design how it looks in the shop
     * @return  a fully loaded designs
     *
     * @throws  ApplicationException  if anything goes wrong
     */
    List<Design> getDesigns(boolean getAllContent) throws ApplicationException;
    
    /**
     * gets list of fully loaded designs in the required artist only if they are in the catalogue.
     *
     * @param   artistUid  the artist to get for
     *
     * @return  a list of fully loaded designs
     *
     * @throws  ApplicationException  if anything goes wrong
     */
    List<Design> getDesignsForArtist(Integer artistUid) throws ApplicationException;

    /**
     * gets list of fully loaded designs matching the required filter, regardless of whether they are in the
     * catalogue or not.
     *
     * @param   filter  the filter to use for matching, or null in which case its ignored.
     *
     * @return  a list of fully loaded designs.
     *
     * @throws  ApplicationException  if anything goes wrong
     */
    List<Design> getDesignsAdmin(String filter) throws ApplicationException;
    
    /**
     * gets list of shallow loaded designs, regardless of whether they are in the
     * catalogue or not.
     *
     * @return  a list of shallow loaded designs with only artists loaded.
     *
     * @throws  ApplicationException  if anything goes wrong
     */
    List<Design> getDesignsAdminLight() throws ApplicationException;

    /**
     * called to show all details of a design, to a customer. regardless of whether its in the catalogue. 
     * useful for linking to, from customers sales orders.
     * 
     * @param designID the design to fetch
     * @return the fully loaded design
     * @throws ApplicationException if anything goes wrong
     */
    Design getDesign(Integer designID) throws ApplicationException;
    
    /**
     * gets list of fully loaded artists matching the required filter.
     *
     * @param   filter  the filter to use for matching, or null in which case its ignored.
     *
     * @return  a list of fully loaded artists
     *
     * @throws  ApplicationException  if anything goes wrong
     */
    List<Artist> getArtistsAdmin(String filter) throws ApplicationException;
    
    /**
     * gets all relevant artists / designs which also have matching tags.
     *
     * @param   searchExpression  DOCUMENT ME!
     *
     * @return  a fully loaded list of gallieries
     *
     * @throws  ApplicationException  if anything goes wrong
     */
    SearchResults search(String searchExpression) throws ApplicationException;

    /**
     * saves or updates the given design. update if UID is not null.
     *
     * @param   design             the design containing the data to save.
     * @return the uid of the saved design
     * @throws  ApplicationException  if anything goes wrong
     *
     * @see     DesignService#saveOrUpdate(uk.co.maxant.gwtdemo.server.po.Design, List, javax.servlet.http.HttpServletRequest)
     */
    Integer saveOrUpdateDesign(Design design) throws ApplicationException;

    /**
     * saves or updates the given artist. update if UID is not null.
     *
     * @param   artist  the artist containing the data to save.
     *
     * @throws  ApplicationException  if anything goes wrong
     *
     * @see     ArtistService#saveOrUpdate(uk.co.maxant.gwtdemo.server.po.Artist)
     */
    Integer saveOrUpdateArtist(Artist artist) throws ApplicationException;
    
	/**
	 * gets the cloud of designs. ie what designs have been viewed the most.
	 *
	 * @throws  ApplicationException  if anything goes wrong
	 *
	 * @see     PartyService#getArtistViewStats()
	 */
	List<Artist> getArtistViewCloud() throws ApplicationException;
	
	/**
	 * gets the pending sales orders. ie what orders are awaiting intervention.
	 *
	 * @throws  ApplicationException  if anything goes wrong
	 *
	 * @see     SalesOrderService#getPendingSalesOrders()
	 */
	List<SalesOrder> getPendingSalesOrders() throws ApplicationException;
	
	/**
	 * gets the discarded sales orders. ie what orders will never be sold. ie stuff that was
	 * in the cart, but the session was discarded.
	 *
	 * @throws  ApplicationException  if anything goes wrong
	 *
	 * @see     SalesOrderService#getPendingSalesOrders()
	 */
	List<SalesOrder> getDiscardedSalesOrders() throws ApplicationException;
	
	/**
	 * gets the problem sales orders. ie what orders are awaiting intervention.
	 *
	 * @throws  ApplicationException  if anything goes wrong
	 *
	 * @see     SalesOrderService#getProblemSalesOrders()
	 */
	List<SalesOrder> getProblemSalesOrders() throws ApplicationException;
	
	/**
	 * searches for sales orders.
	 *
	 * @throws  ApplicationException  if anything goes wrong
	 *
	 * @see     SalesOrderService#searchSalesOrders()
	 */
	List<SalesOrder> searchSalesOrders(String search) throws ApplicationException;
	
	/**
	 * gets the logged in customers sales orders.
	 *
	 * @throws  ApplicationException  if anything goes wrong
	 *
	 * @see     SalesOrderService#getCustomerSalesOrders()
	 */
	List<SalesOrder> getCustomerSalesOrders() throws ApplicationException;
	
	/**
	 * gets the given sales order.
	 *
	 * @throws  ApplicationException  if anything goes wrong
	 *
	 * @see     SalesOrderService#getSalesOrder(Integer)
	 */
	SalesOrder getSalesOrder(String uid) throws ApplicationException;
	
	/**
	 * registers a user. to be called when a user clicks to register so that they can login.
	 *
	 * @return the customer ID.
	 * @throws  ApplicationException  if anything goes wrong
	 *
	 * @see     PartyService#registerUser(uk.co.maxant.gwtdemo.server.po.Users)
	 */
	Integer registerCustomerOrArtist(User user, String password, String captcha, boolean alsoArtist, String artistsDescription) throws ApplicationException;
	
	/**
	 * adds a rating from the logged in user, if they have not already done so.
	 * fails silently if they have already rated the design, or they are not logged in.
	 *
	 * @see     DesignService#rateDesign(Integer, Integer, Integer)
	 */
	void rateDesign(Design design, Integer ratingGiven) throws ApplicationException;

	/**
	 * @param requiredState the state into which this order should be set
	 * @param soUid the uid of the sales order to update
	 * @param notes optional notes to add, if say cancelled.
	 * @throws ApplicationException if anything goes wrong
	 * @see SalesOrderService#updateOrderStatus(uk.co.maxant.gwtdemo.server.po.Users, OrderState, int, String)
	 */
	void updateOrderStatus(OrderState requiredState, int soUid, String notes) throws ApplicationException;
	
	/**
	 * @param requiredState the state into which this order should be set
	 * @param soUid the uid of the sales order to update
	 * @param extraNotes notes that staff want to add
	 * @param txnId the new transaction ID, or null is ok too. a record of the old one is made.
	 * @throws ApplicationException if anything goes wrong
	 * @see SalesOrderService#paymentSuccessful(uk.co.maxant.gwtdemo.server.po.SalesOrder, uk.co.maxant.gwtdemo.server.SalesOrderService.StateUpdater)
	 * @see SalesOrderService#paymentFailed(uk.co.maxant.gwtdemo.server.po.SalesOrder, uk.co.maxant.gwtdemo.server.SalesOrderService.StateUpdater)
	 */
	void updatePaymentStatus(OrderState requiredState, int soUid, String extraNotes, String txnId) throws ApplicationException;

	/**
	 * adds shop notes to the order.
	 * @param soUid
	 * @param notes the notes to add
	 * @throws ApplicationException if anything goes wrong
	 * @see SalesOrderService#addShopNotes(uk.co.maxant.gwtdemo.server.po.Users, int, String)
	 */
	void addShopNotes(int soUid, String notes) throws ApplicationException;

	/**
	 * marks the order as NOT needing attention.
	 * @param soUid
	 * @param msg notes to also be added
	 * @throws ApplicationException if anything goes wrong
	 * @see SalesOrderService#markOrderAsNotNeedingAttention(uk.co.maxant.gwtdemo.server.po.Users, int)
	 */
	void markOrderAsNotNeedingAttention(int soUid, String msg) throws ApplicationException;

	/**
	 * marks the order as NEEDING attention, which also sends a mail if so configured.
	 * @param orderNum
	 * @param msg notes to also be added
	 * @throws ApplicationException if anything goes wrong
	 * @see SalesOrderService#markOrderAsNeedingAttention(uk.co.maxant.gwtdemo.server.po.Party, Integer, String)
	 */
	void markOrderAsNeedingAttention(Integer orderNum, String msg) throws ApplicationException;

	/**
     * must be in the designCommentsModerator role to do this.
     * 
     * gets all designs.
     * 
	 * @see OrchestrationService#getAllComments(Integer)
	 */
	List<DesignComment> getAllComments(Integer designUid) throws ApplicationException;

	/**
	 * gets comments for a given design, which are publically available
	 * 
	 * @see OrchestrationService#getComments(Integer)
	 */
	List<DesignComment> getComments(Integer designUid) throws ApplicationException;

	/**
	 * add a design comment. what a registered, logged in user might want to do.
	 * 
	 * @see OrchestrationService#addDesignComment(DesignComment)
	 */
	Integer addDesignComment(DesignComment dc) throws ApplicationException;
	
	/**
	 * if so configured, a moderator needs to approve a design comment before it goes live.
	 * 
	 * @see OrchestrationService#approveDesignComment(DesignComment)
	 */
	void approveDesignComment(DesignComment dc) throws ApplicationException;
	
	/**
	 * if so configured, a moderator may want to reject a design comment so that it doesnt go live.
	 * 
	 * @see OrchestrationService#rejectDesignComment(DesignComment)
	 */
    void rejectDesignComment(DesignComment dc) throws ApplicationException;

    /**
     * @return a list of designs which need moderation, 
     * 			because moderation is active.
     */
    List<Design> getDesignsPendingModeration() throws ApplicationException;
    
    /**
     * @return the currently logged in user, as an artist. if they are not an artist, we get an error!
     */
    Artist getLoggedInArtist() throws ApplicationException;

    /**
     * updates the given customer/artists profile.
     * 
     * @param artist the user or artist to update.
     * 
     * @throws ApplicationException if anything goes wrong, like duplicate email address or tech error.
     */
    Integer updateCustomer(User user, String newPassword) throws ApplicationException;

    /**
     * used to update an artists profile, when they do it themselves.
     * @param artist object containing artist to update, although the logged 
     * 				in users uid will be used to prevent hacking.
     * @return the uid of the logged in user
     * @throws ApplicationException
     */
    Integer updateArtistProfile(Artist artist) throws ApplicationException;

}
