/*  
 * Copyright (c) 2009 Ant Kutschera, maxant
 * 
 * This file is part of the maxant GWT Demo.
 * 
 * maxant GWT Demo is free software: you can redistribute it and/or modify
 * it under the terms of the Lesser GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * maxant GWT Demo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Lesser GNU General Public License for more details.
 * You should have received a copy of the Lesser GNU General Public License
 * along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */

package uk.co.maxant.gwtdemo.client.generic.widgets;

import com.google.gwt.event.dom.client.KeyUpEvent;
import com.google.gwt.user.client.Timer;

/**
 * a filter component which makes server calls to get the content which should be displayed.
 */
public abstract class AsyncFilterBox extends VeryAbstractFilterBox {

	protected Timer caller;
	
	protected int millisecondsReactionWait;

	protected long lastInput = 0L;
	
	/**
	 * @param labelText
	 * @param infoText
	 * @param resultsText
	 * @param clearTooltip
	 * @param infoTooltip
	 * @param accessKey
	 * @param filterTextboxWidth
	 * @param millisecondsReactionWait waits this long before making a server call. this means that 
	 * 			not every key event is sent to the server.
	 */
	public AsyncFilterBox(
					String labelText, 
					final String infoText, 
					String resultsText, 
					String clearTooltip,
					String infoTooltip,
					char accessKey,
					int filterTextboxWidth,
					int millisecondsReactionWait
					){
		super(labelText, 
				infoText, 
				resultsText, 
				clearTooltip,
				infoTooltip,
				accessKey,
				filterTextboxWidth);
		
		this.millisecondsReactionWait = millisecondsReactionWait;
	}
	
	/**
	 * if event is not null, and its an arrow key, and we are armed (ready to make a server call), then
	 * the time is extended. however we are not armed, then this call simply returns and does not arm.
	 * that is because if you use the arrow keys its annoying if you suddenly end up calling the server
	 * when you dont want to.
	 * 
	 * if fireImmediately is true, then it fires, regardless of the event.
	 * 
	 * @see VeryAbstractFilterBox#doFilter(com.google.gwt.event.dom.client.KeyUpEvent, boolean)
	 */
	@Override
	protected void doFilter(KeyUpEvent event, boolean fireImmediately) {
		
		if(fireImmediately){
			input.setEnabled(false);
			doServerCall(input.getText());
		}else{
			//if not armed, and just an arrow key, then do nothing

			if(event != null){
				if(event.isDownArrow() || event.isUpArrow() || event.isLeftArrow() || event.isRightArrow()){
					//dont need to fire on arrow keys if we are already armed.
					if(caller == null){
						//no armed! no timer is waiting to make the server call!
						return;
					}
				}
			}
			
			//note time. if no timer is running start a new one. 
			lastInput = System.currentTimeMillis();
			schedule_(millisecondsReactionWait);
		}
	}

	/**
	 * if theres not timer, it creates one to run in 
	 */
	private void schedule_(int delay) {
		if(caller == null){
			//no timer - create one
			caller = new Timer(){
				@Override
				public void run() {
					int diff = (int)(System.currentTimeMillis() - lastInput);
					if(diff > millisecondsReactionWait){
						input.setEnabled(false);
						doServerCall(input.getText());
					}else if(diff > 0){
						//schedule another timer for the difference
						caller = null;
						schedule_(diff);
					}else{
						//timer has completely elapsed, so lets start a brand new one
						caller = null;
						schedule_(millisecondsReactionWait);
					}
				}
			};
			caller.schedule(delay);
		}
	}

	/**
	 * called by the component when it is ready to go call the server.
	 */
	protected abstract void doServerCall(String filter);

	/**
	 * must be called by implementing class once the server call returns, otherwise
	 * future calls to the server will not be made.
	 */
	public void serverCallComplete(int numberMatches){
		caller = null;
		results.setText(resultsText + numberMatches);
		input.setEnabled(true);
	}

}

