/*  
 * Copyright (c) 2009 Ant Kutschera, maxant
 * 
 * This file is part of the maxant GWT Demo.
 * 
 * maxant GWT Demo is free software: you can redistribute it and/or modify
 * it under the terms of the Lesser GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * maxant GWT Demo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Lesser GNU General Public License for more details.
 * You should have received a copy of the Lesser GNU General Public License
 * along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */

package uk.co.maxant.gwtdemo.client.to;

/**
 * order statii
 */
public enum OrderState {

	/**
	 * created, but no payment informtion is known. happens just before user gets sent to payment system. its
	 * possible that they cancel from within the payment system and the order is never taken any further, so
	 * from that perspective, an order in this state is not necessarily placed.
	 */
	CREATED(1, "Created"),

	/** the order has been payed for, however payment status is unknown. */
	PLACED(2, "Placed"),

	/** the order needs to be picked. payment may be successful or not, depending upon policies. */
	AWAITING_PICKING(4, "Awaiting picking"),

	/** picking has physically started. */
	STARTED_PICKING(8, "Started picking"),

	/**
	 * picking has been completed. the order is boxed. it may not be ready for dispatch, depending upon
	 * payment status and policies.
	 */
	FINISHED_PICKING(16, "Finished Picking"),

	/** this order has been picked and should be dispatched. */
	READY_FOR_DISPATCH(32, "Ready for dispatch"),

	/** this order has been collected by the delivery partner. */
	DISPATCHED(64, "Dispatched"),

	/** the delivery partner has confirmed that this order was delivered. */
	DELIVERED(128, "Delivered"),

	/** this order has been cancelled. */
	CANCELLED(256, "Cancelled"),
	
	/** not failed and not successful. simply unknown right now. */
	PAYMENT_PENDING(512, "Payment pending"),

	/**
	 * payment has been confirmed by the payment system or staff as being successful. see shop notes for
	 * details.
	 */
	PAYMENT_SUCCESSFUL(1024, "Payment successful"),

	/**
	 * payment has been confirmed by the payment system or staff as having failed. see shop notes for details.
	 */
	PAYMENT_FAILED(2056, "Payment failed"),

	/**
	 * such an order was never created. the user let the session expire before going to the payment system.
	 * this is useful only for statistics.
	 */
	DISCARDED(4096, "Discarded");
	
	/** DOCUMENT ME! */
	private final int value;

	private final String name;
	
	/**
	 * Creates a new OrderState object.
	 *
	 * @param value DOCUMENT ME!
	 */
	private OrderState(int value, String name) {
		this.value = value;
		this.name = name;
	}

	public int getValue() {
		return value;
	}
	
	/**
	 * returns the next available states, or empty, if none are available.
	 */
	public static OrderState[] getNext(OrderState state) {
		//check end states first
		if(state.equals(OrderState.CANCELLED)){
			//or maybe reinstated?
			return new OrderState[0];
		}
		if(state.equals(OrderState.DELIVERED)){
			//nothing more can be done! or maybe cancelled?
			return new OrderState[0];
		}
		if(state.equals(OrderState.DISCARDED)){
			//too late to do owt else
			return new OrderState[0];
		}
		if(state.equals(OrderState.PAYMENT_FAILED)){
			//special case. always allow user to change payment status.
			OrderState[] oss = new OrderState[1];
			oss[0] = OrderState.PAYMENT_SUCCESSFUL;
			return oss;
		}
		if(state.equals(OrderState.PAYMENT_SUCCESSFUL)){
			//special case. always allow user to change payment status.
			OrderState[] oss = new OrderState[1];
			oss[0] = OrderState.PAYMENT_FAILED;
			return oss;
		}
		if(state.equals(OrderState.PAYMENT_PENDING)){
			//special case. allow either payment states
			OrderState[] oss = new OrderState[2];
			oss[0] = OrderState.PAYMENT_SUCCESSFUL;
			oss[1] = OrderState.PAYMENT_FAILED;
			return oss;
		}

		if(state.equals(OrderState.PLACED)){
			//special case. allow either awaiting picking OR cancelled
			OrderState[] oss = new OrderState[2];
			oss[0] = OrderState.AWAITING_PICKING;
			oss[1] = OrderState.CANCELLED;
			return oss;
		}
		
		if(state.equals(OrderState.AWAITING_PICKING)){
			//special case. allow either started picking OR cancelled
			OrderState[] oss = new OrderState[2];
			oss[0] = OrderState.STARTED_PICKING;
			oss[1] = OrderState.CANCELLED;
			return oss;
		}
		
		int i = state.getValue() * 2;
		for(OrderState os : OrderState.values()){
			if(os.getValue() == i){
				OrderState[] oss = new OrderState[1];
				oss[0] = os;
				return oss;
			}
		}
		
		//none others are known!
		return new OrderState[0];
	}
	
	/**
	 * DOCUMENT ME!
	 *
	 * @return DOCUMENT ME!
	 */
	public String getName() {
		return name;
	}

	public static OrderState getOrderState(SalesOrder so){
		if(so.getStatusDiscarded() != null){
			return OrderState.DISCARDED;
		}else if(so.getStatusCancelled() != null){
			return OrderState.CANCELLED;
		}else if(so.getStatusDelivered() != null){
			return OrderState.DELIVERED;
		}else if(so.getStatusDispatched() != null){
			return OrderState.DISPATCHED;
		}else if(so.getStatusReadyForDispatch() != null){
			return OrderState.READY_FOR_DISPATCH;
		}else if(so.getStatusFinishedPicking() != null){
			return OrderState.FINISHED_PICKING;
		}else if(so.getStatusStartedPicking() != null){
			return OrderState.STARTED_PICKING;
		}else if(so.getStatusAwaitingPicking() != null){
			return OrderState.AWAITING_PICKING;
		}else if(so.getStatusOrderPlaced() != null){
			return OrderState.PLACED;
		}else if(so.getStatusOrderCreated() != null){
			return OrderState.CREATED;
		}else {
			throw new IllegalStateException("unable to determine state of SO " + so.getUid());
		}
	}
	
	public static OrderState getPaymentState(SalesOrder so){
		if(so.getStatusPaymentSuccessful() != null){
			if(so.getStatusPaymentFailed() != null){
				//depends which is newer
				if(so.getStatusPaymentFailed().before(so.getStatusPaymentSuccessful())){
					return OrderState.PAYMENT_SUCCESSFUL;
				}else{
					return OrderState.PAYMENT_FAILED;
				}
			}else{
				return OrderState.PAYMENT_SUCCESSFUL;
			}
		}else if(so.getStatusPaymentFailed() != null){
			return OrderState.PAYMENT_FAILED;
		}else {
			return OrderState.PAYMENT_PENDING;
		}
	}
	


}
