/*  
 * Copyright (c) 2009 Ant Kutschera, maxant
 * 
 * This file is part of the maxant GWT Demo.
 * 
 * maxant GWT Demo is free software: you can redistribute it and/or modify
 * it under the terms of the Lesser GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * maxant GWT Demo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Lesser GNU General Public License for more details.
 * You should have received a copy of the Lesser GNU General Public License
 * along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */
package uk.co.maxant.gwtdemo.server;

import java.math.MathContext;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import javax.servlet.http.HttpServletRequest;

import org.hibernate.SessionFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.orm.hibernate3.HibernateTemplate;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Isolation;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;

import uk.co.maxant.gwtdemo.client.ApplicationException;
import uk.co.maxant.gwtdemo.client.Roles;
import uk.co.maxant.gwtdemo.server.po.Design;
import uk.co.maxant.gwtdemo.server.po.DesignComment;
import uk.co.maxant.gwtdemo.server.po.DesignRating;
import uk.co.maxant.gwtdemo.server.po.DesignTag;
import uk.co.maxant.gwtdemo.server.po.FrontPageDesign;
import uk.co.maxant.gwtdemo.server.po.Party;
import uk.co.maxant.gwtdemo.util.StringUtils;

/**
 * @author maxant
 * @copyright 2009 maxant
 */
@Service("designService")
@Transactional(
		propagation=Propagation.REQUIRED, 
		isolation=Isolation.DEFAULT, 
		rollbackFor=ApplicationException.class)
public class DesignService {

	
	private static final int MAX_DESIGNS_TO_SHOW = 5; //compare to what GUI displays in DesignAdmin

	private static final long ONE_HOUR = 1000L * 3600;

	private HibernateTemplate hibernateTemplate;
	
	private List<Design> featuredDesignCache = new ArrayList<Design>();

	/** when was the last read on the DB? */
	private long lastFeaturedDesignRead = 0L;
	
	private List<Design> newestDesignCache = new ArrayList<Design>();
	
	/** when was the last read on the DB? */
	private long lastNewestDesignRead = 0L;

	@Autowired
	public void setSessionFactory(SessionFactory sessionFactory) {
		this.hibernateTemplate = new HibernateTemplate(sessionFactory);
	}

	@Autowired
	private ConfigService configService;
	
	/**
	 * @return a list of shallow loaded designs for the front page.
	 * @throws ApplicationException if anything goes wrong
	 */
	@SuppressWarnings("unchecked")
	public List<Design> getFeaturedDesigns() throws ApplicationException {
		if(System.currentTimeMillis()-lastFeaturedDesignRead > ONE_HOUR){
			featuredDesignCache.clear();
			List<FrontPageDesign> fpds = this.hibernateTemplate.find(
					"from FrontPageDesign fpd where " +
					"fpd.design.approved = true and " +
					"fpd.design.inCatalogue = true");
			
			for(FrontPageDesign fpd : fpds){
				Design d = fpd.getDesign();
				featuredDesignCache.add(d);
			}
		}
		return featuredDesignCache;
	}

	/**
	 * @return a list of most recently added designs
	 * @throws ApplicationException if anything goes wrong
	 */
	@SuppressWarnings("unchecked")
	public List<Design> getNewestDesigns() throws ApplicationException {
		if(System.currentTimeMillis()-lastNewestDesignRead > ONE_HOUR){
			newestDesignCache.clear();
			List<Design> ds = this.hibernateTemplate.find(
					"from Design d where " +
					"d.approved = true and " +
					"d.inCatalogue = true order by d.created desc");
			
			//TODO 001 optimise, so we dont load the entire db

			for(Design d : ds){
				if(newestDesignCache.size() >= 6) break; //TODO magic number!
				newestDesignCache.add(d);
			}
		}
		return newestDesignCache;
	}
	
	/**
	 * get fully loaded designs for the given artist
	 * @param dontCheckIfInCatalogueOrApproved if true, then gets designs regardless of whether in catalogue or not and regardless whether they are approved or not. otherwise, only loads designs in the catalogue which are also approved.
	 * @return a list of fully loaded designs to be displayed in the shop
	 * @throws ApplicationException in case of errors
	 */
	@SuppressWarnings("unchecked")
	public List<Design> getDesigns(boolean dontCheckIfInCatalogueOrApproved) throws ApplicationException {
		List<Design> designs = new ArrayList<Design>();
		List<Design> ds = this.hibernateTemplate.find("from Design d");
		
		for(Design d : ds){
			if(d.isInCatalogue() || dontCheckIfInCatalogueOrApproved){
				if(d.isApproved() || dontCheckIfInCatalogueOrApproved){
					designs.add(d);
				}
			}
		}
		return designs;
	}

	/**
	 * get fully loaded designs for the given artist
	 * @param artistUid artist to load from
	 * @param dontCheckIfInCatalogueOrApproved if true, then gets designs regardless of whether in catalogue or not and regardless whether they are approved or not. otherwise, only loads designs in the catalogue which are also approved.
	 * @return a list of fully loaded designs to be displayed in the shop, for the given artist
	 * @throws ApplicationException in case of errors
	 */
	public List<Design> getDesignsForArtist(Integer artistUid, boolean dontCheckIfInCatalogueOrApproved) throws ApplicationException {
		List<Design> designs = new ArrayList<Design>();
		Party user = (Party) this.hibernateTemplate.load(Party.class, artistUid);
		
		for(Design d : user.getDesigns()){
			if(d.isInCatalogue() || dontCheckIfInCatalogueOrApproved){
				if(d.isApproved() || dontCheckIfInCatalogueOrApproved){
					designs.add(d);
				}
			}
		}
		return designs;
	}
	
	/**
	 * adds a rating from the logged in user, if they have not already done so.
	 * fails silently if they have already rated the design, or they are not logged in.
	 * 
	 * @param 
	 * @throws ApplicationException in case of errors
	 */
	@SuppressWarnings("unchecked")
	public void rateDesign(Integer designUid, Integer userUid, Integer ratingGiven) throws ApplicationException {
		if(userUid == null){
			//cannot add a rating unless user is logged in
			return;
		}
		
		
		List<DesignRating> drs = this.hibernateTemplate.find("from DesignRating dr where dr.design.uid = ? and dr.party.uid = ?", new Object[]{designUid, userUid});
		if(drs.size() != 0){
			//only one rating per user!
			return;
		}

		//use the latest object out the DB...
		Design d = (Design)this.hibernateTemplate.load(Design.class, designUid);
		d.setNumberOfRatings(d.getNumberOfRatings() + 1);
		d.setSumOfRatings(d.getSumOfRatings() + ratingGiven);

		//now add a design rating
		DesignRating dr = new DesignRating();
		dr.setDesign(d);
		Party user = new Party();
		user.setUid(userUid);
		dr.setParty(user);
		dr.setVal(ratingGiven);
		this.hibernateTemplate.save(dr);
	}

	/**
	 * get fully loaded designs matching the filter, regardless whether they are in the catalogue.
	 * @param filter filter to compare against. or null, in which case everything matches. more efficient that way too.
	 * @return a list of fully loaded designs to be displayed eg in design admin
	 * @throws ApplicationException in case of errors
	 */
	@SuppressWarnings("unchecked")
	public List<Design> getDesigns(String filter) throws ApplicationException {
		List<Design> designs = new ArrayList<Design>();
		List<Design> ds = this.hibernateTemplate.find("from Design d");

		if(filter != null){
			filter = filter.trim().toLowerCase();
			filter = ".*" + filter + ".*";
		}
		
		for(Design d : ds){
				
			boolean matches = false;
			if(filter == null){
				
				//show regardless. and also dont break, we need to really return ALL designs in this case
				
				matches = true;
			}else {
				
				//if we have enough designs, break
				if(designs.size() > MAX_DESIGNS_TO_SHOW){
					break;
				}
				
				List<String> stringsToCompare = new ArrayList<String>();
				
				for(DesignTag dt : d.getDesignTags()){
					stringsToCompare.add(dt.getTag().toLowerCase());
				}
				stringsToCompare.add(String.valueOf(d.getUid()));
				stringsToCompare.add(String.valueOf(d.getName().toLowerCase()));
				stringsToCompare.add(String.valueOf(d.getDescription().toLowerCase()));

				for(String s : stringsToCompare){
					if(s.toLowerCase().matches(filter)){
						matches = true;
						break;
					}
				}
			}
			if(matches){
				designs.add(d);
			}
		}
		return designs;
	}
	
	/**
	 * saves/updates the design object. saves all tags as new ones and deletes all existing ones.
	 * 
	 * @return the uid of the saved object
	 * 
	 * must be in "backoffice" role.
	 */
	public Integer saveOrUpdate(Design d, List<String> tags, HttpServletRequest request) throws ApplicationException {
		MathContext mc = new MathContext(12);

		Design design = null;
		if(d.getUid() == null){
			d.setBasePrice(d.getBasePrice().round(mc)); //check before saving!
			d.setCreated(new Date());
			Integer uid = (Integer)this.hibernateTemplate.save(d);
			design = d;
			design.setUid(uid);
		}else{
			design = (Design)this.hibernateTemplate.load(Design.class, d.getUid());
			design.setName(d.getName());
			design.setDescription(d.getDescription());
			design.setInCatalogue(d.isInCatalogue());
			design.setParty(d.getParty());
			design.setBasePrice(d.getBasePrice().round(mc)); //this one needs to be checked in all cases!

			//only someone in backoffice may modify this!
			if(d.isApproved() != design.isApproved()){
				if(!request.isUserInRole(Roles.BACKOFFICE)){
					throw new SecurityException("only backoffice may modify the approval status of a design");
				}
			}
			design.setApproved(d.isApproved());
		}

		if(tags != null){
			//remove superfluous tags
			List<String> alreadyExistingTags = new ArrayList<String>();
			for(DesignTag tag : design.getDesignTags()){
				if(!tags.contains(tag.getTag())){
					this.hibernateTemplate.delete(tag);
				}else{
					alreadyExistingTags.add(tag.getTag());
				}
			}
		
			//add newly added tags
			for(String tag : tags){
				if(!alreadyExistingTags.contains(tag)){
					DesignTag dt = new DesignTag(design, tag);
					this.hibernateTemplate.save(dt);
				}
			}
		}

		return design.getUid();
	}
	
	/**
	 * gets the given design.
	 */
	public Design getDesign(Integer uid) throws ApplicationException {
		Design design = (Design)this.hibernateTemplate.load(Design.class, uid);

		return design;
	}

	/**
	 * @return gets ALL comments. meant to be called by moderators.
	 */
	public List<DesignComment> getAllComments(Integer designUid) throws ApplicationException {
		@SuppressWarnings("unchecked")
		List<DesignComment> dcs = this.hibernateTemplate.find("from DesignComment dc where dc.design.uid = ? order by dc.uid desc", designUid);

		return dcs;
	}

	/**
	 * if moderation is configured, then only shows comments which are approved and not rejected.
	 * 
	 * if moderation is NOT configured, then shows all comments for the design.
	 * 
	 * @return gets comments. meant to be called by any user. provides the
	 * 					comments which should be displayed in the shop.
	 */
	@SuppressWarnings("unchecked")
	public List<DesignComment> getComments(Integer designUid) throws ApplicationException {
		boolean moderationIsEnabled = configService.boolValue(ConfigService.designCommentModerationRequired);

		String sql = "from DesignComment dc where dc.design.uid = ?";
		if(moderationIsEnabled){
			sql += " and dc.approved is not null and dc.rejected is null";
		}
		sql += " order by dc.uid desc";
		List<DesignComment> dcs = this.hibernateTemplate.find(sql, designUid);
		
		return dcs;
	}

	/**
	 * saves/updates the design comment object. 
	 * 
	 * must be in "designCommentsModerator" role.
	 * 
	 * @return the uid of the object
	 */
	public Integer saveOrUpdate(DesignComment dc) throws ApplicationException {
		DesignComment designComment = null;
		if(dc.getUid() == null){
			//ensure no html / script
			dc.setComment(StringUtils.stripHTML(dc.getComment()));
			Integer uid = (Integer)this.hibernateTemplate.save(dc);
			designComment = dc;
			designComment.setUid(uid);
		}else{
			designComment = (DesignComment)this.hibernateTemplate.load(DesignComment.class, dc.getUid());
			designComment.setComment(StringUtils.stripHTML(dc.getComment()));
			designComment.setApproved(dc.getApproved());
			designComment.setComment(dc.getComment());
			designComment.setRejected(dc.getRejected());
			designComment.setPartyByModeratedByUid(dc.getPartyByModeratedByUid());
		}

		return designComment.getUid();
	}
	
	/**
	 * if moderation is configured, then only shows comments which are approved and not rejected.
	 * 
	 * if moderation is NOT configured, then shows all comments for the design.
	 * 
	 * @return gets comments. meant to be called by any user. provides the
	 * 					comments which should be displayed in the shop.
	 */
	@SuppressWarnings("unchecked")
	public List<Design> getDesignsPendingModeration() throws ApplicationException {
		boolean moderationIsEnabled = configService.boolValue(ConfigService.designCommentModerationRequired);
		
		if(!moderationIsEnabled){
			throw new ApplicationException("DESS014", "designCommentModerationRequired=false", "Moderation is not active. There is nothing for you to do here.");
		}

		String sql = "from DesignComment dc where dc.approved is null and dc.rejected is null order by dc.uid desc";
		List<DesignComment> dcs = this.hibernateTemplate.find(sql);
		
		List<Design> ds = new ArrayList<Design>();
		for(DesignComment dc : dcs){
			ds.add(dc.getDesign());
		}
		return ds;
	}

	/**
	 * does a search on design tags as well as objects.
	 * 
	 * @return a list of deeply loaded designs and artists.
	 * @throws ApplicationException in case of errors
	 */
	@SuppressWarnings("unchecked")
	public SearchResult search(String searchExpression) throws ApplicationException {

		List<Design> designs = new ArrayList<Design>();
		List<Integer> designIds = new ArrayList<Integer>(); //no hashCode, so cannot call "contains"
		List<Party> artists = new ArrayList<Party>();
		List<Integer> artistIds = new ArrayList<Integer>(); //no hashCode, so cannot call "contains"
		
		Integer uid = 0;
		try{
			uid = Integer.parseInt(searchExpression);
		}catch(NumberFormatException nfe){
			//oh well, not a number, so skip UID searching
		}

		if(searchExpression == null){
			searchExpression = "%"; //full wildcard search!
		}
		
		//do a wildcard search
		searchExpression = "%" + searchExpression + "%";
		
		//////////////////////////
		//
		// DESIGNS
		//
		//////////////////////////
		
		//now search for design tags
		Object[] params = new Object[]{searchExpression};
		List<DesignTag> dts = this.hibernateTemplate.find(
				"from DesignTag dt " +
				"where " +
				"dt.design.inCatalogue = true and " +
				"dt.design.approved = true and " +
				"dt.tag like ?", params );
		for(DesignTag dt : dts){
			if(designIds.size() > 100) break; //dont collect the whole world - memory!!
			if(!designIds.contains(dt.getDesign().getUid())){ //not already in there
				designs.add(dt.getDesign());
				designIds.add(dt.getDesign().getUid());
			}
		}
		
		//now search for design attributes...
		params = new Object[3];
		int i = 0;
		params[i++] = searchExpression;
		params[i++] = searchExpression;
		params[i++] = uid;
		List<Design> ds = this.hibernateTemplate.find(
				"from Design d " +
				"where " +
				"(d.name like ? or " +
				"d.description like ? or " +
				"d.uid like ?) " +
				"AND d.inCatalogue = true " +
				"AND d.approved = true", params );
		for(Design d : ds){
			if(designIds.size() > 100) break; //dont collect the whole world - memory!!
			if(!designIds.contains(d.getUid())){ //not already in there
				//all criteria are met :-)
				designs.add(d);
				designIds.add(d.getUid());
			}
		}

		
		//////////////////////////
		//
		// ARTISTS
		//
		//////////////////////////
		//now search for artist attributes...
		params = new Object[3];
		i = 0;
		params[i++] = searchExpression;
		params[i++] = searchExpression;
		params[i++] = uid;
		List<Party> us = this.hibernateTemplate.find(
				"from Party u where " +
				"(u.screenname like ? or " +
				"u.artistDescription like ? or " +
				"u.uid like ? ) " +
				"AND u.isArtist = true", params );
		
		for(Party u : us){
			if(artistIds.size() > 100) break; //dont collect the whole world - memory!!
			if(!artistIds.contains(u.getUid())){ //not already in there
				//do they have any designs that customers can buy??
				for(Design d : u.getDesigns()){
					if(d.isApproved() && d.isInCatalogue()){
						artists.add(u);
						artistIds.add(u.getUid());
						break;
					}
				}
			}
		}

		//TODO for searching, how to order results?? still to specify...

		return new SearchResult(designs, artists);
	}

	static class SearchResult {
		List<Design> designs;
		List<Party> artists;

		public SearchResult(List<Design> designs, List<Party> artists) {
			this.designs = designs;
			this.artists = artists;
		}
	}

}
