/*  
 * Copyright (c) 2009 Ant Kutschera, maxant
 * 
 * This file is part of the maxant GWT Demo.
 * 
 * maxant GWT Demo is free software: you can redistribute it and/or modify
 * it under the terms of the Lesser GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * maxant GWT Demo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Lesser GNU General Public License for more details.
 * You should have received a copy of the Lesser GNU General Public License
 * along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */

package uk.co.maxant.gwtdemo.server;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.Date;
import java.util.StringTokenizer;

import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Isolation;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;

import uk.co.maxant.gwtdemo.client.ApplicationException;
import uk.co.maxant.gwtdemo.server.util.LogHelper;
import uk.co.maxant.gwtdemo.server.util.ServerFormats;


/**
 * The point of this class is to log errors so that they are brought to the attention of someone who
 * can deal with them. This is done by first logging the error, then writing the error to the database,
 * and finally by sending an email. If any of these fail, the others are still attempted. For each failure,
 * a message is also put in the log.
 */
@Service("errorService")
@Transactional(propagation=Propagation.REQUIRED, isolation=Isolation.DEFAULT, rollbackFor=ApplicationException.class)
public class ErrorService {
	
	private static Logger log = LogHelper.getLog();

	@Autowired
	private MailService mailService;

	@Autowired
	private ConfigService configService;

	/**
	 * sends a mail, unless that fails in which case it puts a large log entry in the log.
	 */
	public String reportError(String appName, String description, Throwable t){
		
		boolean failedMailSend = false;
		
		Date now = new Date();
		String uid = ServerFormats.yyyyMMddHHmmssSSS().format(now);
		
		//log the error
		log.error("An error occurred (UID/timestamp is " + uid + ") in application '" + appName + "'. The error was: " + description, t);
		
		//send mail
		try{
			String msg = "An error has occurred in '" + appName + "'. The UID/timestamp of the error is " + 
				uid + "<br><br>" + 
				"Please use the logs for this application to find the root cause of the error and provide a " +
				"suitable solution.<br><br>The description is: " + description + "<br><br>" +
				"The error was: " + getStackTrace(t, "<br>");

			String addy = configService.value(ConfigService.emailAdmin);
			mailService.sendMail(
					addy, 
					"ERROR", 
					msg);

		}catch(ApplicationException e){
			failedMailSend = true;
			log.error("Failed to send email to report the above error", e);
		}
			
		if(failedMailSend){
			//this is a fatal condition because no one will know about it!
			//Lets write to the log file with the word PANIC so we can easily search for it daily...
			log.error(  "\r\n\r\n******************************\r\n" +
						"* PANIC!                     *\r\n" +
						"* =======                    *\r\n" +
						"* (DF29E)                    *\r\n" +
						"* Failed to log the above    *\r\n" +
						"* error to both the database *\r\n" +
						"* and email. No one has been *\r\n" +
						"* informed!!                 *\r\n" +
						"******************************\r\n\r\n");
		}
		
		return uid;
	}

	/**
	 * Get the stack trace as a string, delimited with \r and \n.
	 */
	public static String getStackTrace(Throwable t){
		return getStackTrace(t, null);
	}
	
	/**
	 * Get the stack trace as a string, delimited with the given charater.
	 */
	public static String getStackTrace(Throwable t, String delimiter){
		StringWriter sw = new StringWriter();
		PrintWriter pw = new PrintWriter(sw);
		t.printStackTrace(pw);
		String stackTrace = sw.toString();
		
		if(delimiter != null){
			//replace all \r\n with the delimiter
			StringBuilder newStackTrace = new StringBuilder();
			StringTokenizer st = new StringTokenizer(stackTrace, "\r\n");
			while(st.hasMoreTokens()){
				newStackTrace.append(st.nextToken());
				if(st.hasMoreTokens()){
					newStackTrace.append(delimiter);
				}
			}
			stackTrace = newStackTrace.toString();
		}
		return stackTrace;
	}
	
}
