/*  
 * Copyright (c) 2009 Ant Kutschera, maxant
 * 
 * This file is part of the maxant GWT Demo.
 * 
 * maxant GWT Demo is free software: you can redistribute it and/or modify
 * it under the terms of the Lesser GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * maxant GWT Demo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Lesser GNU General Public License for more details.
 * You should have received a copy of the Lesser GNU General Public License
 * along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */

package uk.co.maxant.gwtdemo.server;

import java.io.File;
import java.io.FilenameFilter;
import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.security.Principal;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpSession;

import org.apache.log4j.Logger;
import org.hibernate.SessionFactory;
import org.hibernate.stat.Statistics;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.orm.hibernate3.HibernateTemplate;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Isolation;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;

import uk.co.maxant.gwtdemo.client.ApplicationException;
import uk.co.maxant.gwtdemo.client.ApplicationLayerService;
import uk.co.maxant.gwtdemo.client.Roles;
import uk.co.maxant.gwtdemo.client.to.Artist;
import uk.co.maxant.gwtdemo.client.to.BasketItem;
import uk.co.maxant.gwtdemo.client.to.Design;
import uk.co.maxant.gwtdemo.client.to.DesignComment;
import uk.co.maxant.gwtdemo.client.to.OrderState;
import uk.co.maxant.gwtdemo.client.to.PaypalConfig;
import uk.co.maxant.gwtdemo.client.to.SalesOrder;
import uk.co.maxant.gwtdemo.client.to.SearchResults;
import uk.co.maxant.gwtdemo.client.to.SessionInfo;
import uk.co.maxant.gwtdemo.client.to.User;
import uk.co.maxant.gwtdemo.server.PartyService.ArtistViewStat;
import uk.co.maxant.gwtdemo.server.po.Party;
import uk.co.maxant.gwtdemo.server.util.LogHelper;
import uk.co.maxant.gwtdemo.server.util.Mapper;
import uk.co.maxant.gwtdemo.server.util.SalesOrderHelper;
import uk.co.maxant.spring.security.RolesAllowed;


/**
 * the top level facade service which lives inside spring, so that it starts and commits transactions. the
 * application layer calls this service directly without containing any functionality to ensure
 * transactionality and security is guaranteed.
 */
@Service("orchestrationService")
@Transactional(
    propagation = Propagation.REQUIRED,
    isolation = Isolation.DEFAULT, 
    rollbackFor=ApplicationException.class
)
public class OrchestrationService extends Mapper {
	
	private static final int MAX_NUMBER_OF_ARTISTS_TO_RETURN = 30;
	private static final int MAX_NUMBER_OF_DESIGNS_TO_RETURN = 30;

    private static final long serialVersionUID = 1L;

    private static final Logger log = LogHelper.getLog();
    
    /** constant for accessing the contents of the basket out of the web session. */
    public static final String BASKET = "_BASKET";

	public static final String USER = "_USER";

    @Autowired
    private DesignService designService;

    @Autowired
    private SalesOrderService salesOrderService;

    @Autowired
    private PartyService userService;

    @Autowired
    private ConfigService configService;

	@Autowired
	protected MailService mailService;

	private HibernateTemplate hibernateTemplate;

	private Stats lastStats = new Stats();
	
	@Autowired
	public void setSessionFactory(SessionFactory sessionFactory) {
		this.hibernateTemplate = new HibernateTemplate(sessionFactory);
	}

	private String path = null;

    /**
     * @see  ApplicationLayerService#getSessionInfo()
     */
    public SessionInfo getSessionInfo(HttpServletRequest request) throws ApplicationException {
    	
    	try{
    		if(path == null){
    			path = request.getSession().getServletContext().getRealPath("/index.jsp");
    			path = path.substring(0, path.length() - 10);
    		}
        	
            SessionInfo info = new SessionInfo();
            info.setSessionId(request.getSession().getId());

            Party u = userService.getLoggedinUser(request.getUserPrincipal());

            if (u != null) {
                info.setLogin(u.getLogin());
                info.setUserUid(u.getUid());
                info.setFirstname(u.getFirstname());
                info.setLastname(u.getLastname());

                for (String role : userService.getRoles(u)) {
                    info.getRoles().add(role);
                }

                if (request.isUserInRole(Roles.BACKOFFICE) || request.isUserInRole(Roles.PICKING)) {

                    // add masterData
                    List<Party> artists = userService.getAllArtists();

                    for (Party artist : artists) {
                        Artist a = mapArtist(artist);
                        info.getAllArtists().add(a);
                    }
                }
            }

        	info.setSpotlights(getSpotlights());
        	
        	info.setDesignCommentModerationActive(configService.boolValue(ConfigService.designCommentModerationRequired));
            
            return info;
    	}catch(Exception e){
    		handleException("ORCH001", "failed to get session info: " + e.getMessage(), e);
    	}
    	return null; //never happens since #handleException() ALWAYS throws one
    }

    /**
     * logs exception as it passes out of the spring context 
     * (ie this service) and transforms it into an application 
     * exception to ensure a rollback. regardless of what happens 
     * we want a rollback. spring only automatically performs a 
     * rollback if the exception is runtime. but regardless, this
     * application requires that any exception leads to a rollback.
     */
	private void handleException(String code, String msg, 
						Exception e) throws ApplicationException {

		log.error(msg, e);
		
		if(e instanceof ApplicationException){
			throw (ApplicationException)e;
		}
	
		throw new ApplicationException(
				code, 
				msg + "\r\n" + e.getMessage(), 
				"An error occurred. Please try again.");
	}

	/**
     * @return the list of images in the spotlights folder
     */
    private String[] getSpotlights() {
        //finally set the number of spotlights we have, so that the browser can load them randomly
        File f = new File(path, "images/spotlight");
        String[] files = f.list(new FilenameFilter() {
			public boolean accept(File dir, String name) {
				//MUST end with jpg, gif or png
				if(name.toLowerCase().endsWith(".jpg")){
					return true;
				}else if(name.toLowerCase().endsWith(".gif")){
					return true;
				}else if(name.toLowerCase().endsWith(".png")){
					return true;
				}
				return false;
			}
		});
        if(files == null){
        	//not ideal...
        	return new String[]{};
        }else{
        	return files;
        }
	}

	/**
     * @see  ApplicationLayerService#addToBasket(BasketItem)
     */
    public List<BasketItem> addToBasket(BasketItem basketItem, 
    		HttpServletRequest request) throws ApplicationException {

    	try{
	    	List<BasketItem> basket = getBasket(request);
	        basket.add(basketItem);
	
	        return basket;
		}catch(Exception e){
			handleException(
					"ORCH002",
					"failed to get session info",
					e);
		}
		return null; //never happens since #handleException() ALWAYS throws one
    }

    /**
     * @see  ApplicationLayerService#emptyBasket()
     */
    public List<BasketItem> emptyBasket(HttpServletRequest request) throws ApplicationException {

    	try{
    		List<BasketItem> basket = getBasket(request);
    		basket.clear();
    		
    		return basket;
		}catch(Exception e){
			handleException("ORCH003", "failed to get session info: " + e.getMessage(), e);
		}
		return null; //never happens since #handleException() ALWAYS throws one
    }

    /**
     * for use when a user is logged in and a request exists.
     * @see  ApplicationLayerService#createOrder(Address)
     */
    public Integer createOrder(HttpServletRequest request, String email, String name) throws ApplicationException {

    	try{
	    	uk.co.maxant.gwtdemo.server.po.SalesOrder so = createOrder(
	    			request.getUserPrincipal(), 
	    			getBasket(request), 
	    			false,
	    			email, 
	    			name);
	        // store order in session, for paypal redirection page which needs info out of it.
	        log.info("created sales order " + so.getUid() + " for session " + request.getSession().getId());
	        return so.getUid();
		}catch(Exception e){
			handleException("ORCH004", "failed to get session info: " + e.getMessage(), e);
		}
		return null; //never happens since #handleException() ALWAYS throws one
    }
    
    /**
     * for use by session listener, for creating empty orders when session expire.
     * @see  ApplicationLayerService#createOrder(Address)
     */
	public void createDiscardedOrder(HttpSession session) throws ApplicationException {

		try{
	    	List<BasketItem> basket = getBasket(session);
	    	
	    	if(basket == null || basket.size() == 0){
	    		//no need to create a discarded order, since the basket is empty!
	    		return;
	    	}
	    	
	    	Principal principal = (Principal)session.getAttribute(OrchestrationService.USER);
	
	    	createOrder(principal, basket, true, null, null);
		}catch(Exception e){
			handleException("ORCH005", "failed to get session info: " + e.getMessage(), e);
		}
    }

    private uk.co.maxant.gwtdemo.server.po.SalesOrder createOrder(Principal principal, 
    		List<BasketItem> basket, boolean discarded, String email, String name) throws ApplicationException {

    	try{
	    	Party user = userService.getLoggedinUser(principal);
	
	        if(user == null){
	        	//well, we need to create one! see if the user with the given email address already exists...
	        	user = userService.getUser(email);
	        	if(user == null){
	        		//they dont exist in the DB, so lets create one!
	        		user = userService.createAnonymousUser(email, name);
	        	}
	        }
	        
	        List<uk.co.maxant.gwtdemo.server.util.BasketItem> bis =
	            new ArrayList<uk.co.maxant.gwtdemo.server.util.BasketItem>();
	
	        for (BasketItem bi : basket) {
	            bis.add(new uk.co.maxant.gwtdemo.server.util.BasketItem(mapDesign(bi.getDesign())));
	        }
	
	        uk.co.maxant.gwtdemo.server.po.SalesOrder so = salesOrderService.createOrder(bis, user, discarded);
	
	        return so;
		}catch(Exception e){
			handleException("ORCH006", "failed to get session info: " + e.getMessage(), e);
		}
		return null; //never happens since #handleException() ALWAYS throws one
    }

    /**
     * @return  never null. a list of designs in the basket.
     *
     * @see     ApplicationLayerService#getBasket() gets the designs out of the session. if the session
     *          contains nowt, then the list will be created and set in the session.
     */
    public List<BasketItem> getBasket(HttpServletRequest request) throws ApplicationException {

    	try{
	    	List<BasketItem> basket = getBasket(request.getSession());
	    	
	        //in all cases, need to ensure that the user are also stored in the session, in case 
	        //this basket gets discarded and we need to save it off for stats.
	        if(request.getUserPrincipal() != null){
	        	request.getSession().setAttribute(USER, request.getUserPrincipal());
	        }else{
	        	request.getSession().setAttribute(USER, null);
	        }
	
	        return basket;
		}catch(Exception e){
			handleException("ORCH007", "failed to get session info: " + e.getMessage(), e);
		}
		return null; //never happens since #handleException() ALWAYS throws one
    }

	private List<BasketItem> getBasket(HttpSession session) throws ApplicationException {
		@SuppressWarnings("unchecked")
        List<BasketItem> basket = (List<BasketItem>) session.getAttribute(BASKET);
        if (basket == null) {
            basket = new ArrayList<BasketItem>();
            session.setAttribute(BASKET, basket);
        }
        
        return basket;
    }

    /**
     * @see  ApplicationLayerService#getFeaturedDesigns()
     */
    public List<Design> getFeaturedDesigns(HttpServletRequest request) throws ApplicationException {
        try {
            List<uk.co.maxant.gwtdemo.server.po.Design> designs = designService.getFeaturedDesigns();
    		List<Design> ds = mapAndBuildDesigns(designs);

    		Collections.shuffle(ds);
            
    		return ds;
		}catch(Exception e){
			handleException("ORCH008", "failed to get session info: " + e.getMessage(), e);
		}
		return null; //never happens since #handleException() ALWAYS throws one
    }

    /**
     * @see  ApplicationLayerService#getNewestDesigns()
     */
    public List<Design> getNewestDesigns(HttpServletRequest request) throws ApplicationException {
    	try {
    		List<uk.co.maxant.gwtdemo.server.po.Design> designs = designService.getNewestDesigns();
    		List<Design> ds = mapAndBuildDesigns(designs);
    		return ds;
		}catch(Exception e){
			handleException("ORCH009", "failed to get session info: " + e.getMessage(), e);
		}
		return null; //never happens since #handleException() ALWAYS throws one
    }
    
	/**
     * @see  ApplicationLayerService#getPaypalConfig(Integer soUid)
     */
    public PaypalConfig getPaypalConfig(Integer soUid, HttpServletRequest request)
        throws ApplicationException {

        try {
            uk.co.maxant.gwtdemo.server.po.SalesOrder so = salesOrderService.getSalesOrder(soUid);

            if (so == null) {
                throw new IllegalStateException("unknown order " + soUid);
            }

            PaypalConfig pc = new PaypalConfig();
            pc.setPaypalMerchantEmail(configService.value(ConfigService.paypal_merchant_email));
            pc.setPaypalURL(configService.value(ConfigService.paypal_payment_url));
            pc.setItemName("Order " + soUid);
            pc.setOrderUid(soUid.toString());
            pc.setCancelReturnURL(configService.value(ConfigService.paypalCancelReturnURL));
            pc.setEmail(so.getEmail());
            pc.setContinueButtonText("Complete the order");
            pc.setLogo("images/logo.jpg");
            pc.setAllowShippingAddress("1"); // 0=customer is prompted, 1=not asked, 2=required
            pc.setReturnURL(configService.value(ConfigService.paypalReturnURL));
            pc.setIpnURL(configService.value(ConfigService.paypalIPNURL));
            pc.setPaypalCurrency(configService.value(ConfigService.paypal_currency));
            pc.setLanguage(""); // TODO if we leave it empty, does it take the browser lang?
            pc.setTotal(String.valueOf(SalesOrderHelper.getSubTotal(so)));
            pc.setShipping(String.valueOf(so.getDeliveryTotal()));

            return pc;
		}catch(Exception e){
			handleException("ORCH010", "failed to get session info: " + e.getMessage(), e);
		}
		return null; //never happens since #handleException() ALWAYS throws one
    }

    /**
     * @see  ApplicationLayerService#getDesigns(Integer, Integer, boolean)
     */
    public List<Design> getDesigns(boolean getAllContent, HttpServletRequest request)
        throws ApplicationException {

        try {
        	List<uk.co.maxant.gwtdemo.server.po.Design> designs = designService.getDesigns(false);
            List<Design> ds = new ArrayList<Design>();

            for (uk.co.maxant.gwtdemo.server.po.Design design : designs) {
    			Design d = mapDesign(design);
    			d.setArtist(mapArtist(design.getParty()));
    			d.setNumberOfComments(design.getDesignComments().size());
    			ds.add(d);
    			
                userService.incrementArtistViews(design.getParty().getUid());
            }

            return ds;
		}catch(Exception e){
			handleException("ORCH011", "failed to get session info: " + e.getMessage(), e);
		}
		return null; //never happens since #handleException() ALWAYS throws one
    }
    
    /**
     * @see  ApplicationLayerService#getDesignsForArtist(Integer)
     */
    public List<Design> getDesignsForArtist(Integer artistUid, HttpServletRequest request)
    throws ApplicationException {
    	
    	try {
    		List<uk.co.maxant.gwtdemo.server.po.Design> designs = 
    			designService.getDesignsForArtist(
    					artistUid, 
    					false);
    		
    		List<Design> ds = new ArrayList<Design>();

    		for (uk.co.maxant.gwtdemo.server.po.Design design : designs) {
    			
				Design d = mapDesign(design);
				d.setArtist(mapArtist(design.getParty()));
    			d.setNumberOfComments(design.getDesignComments().size());
				ds.add(d);
    		}

            userService.incrementArtistViews(artistUid);

    		return ds;
		}catch(Exception e){
			handleException("ORCH012", "failed to get session info: " + e.getMessage(), e);
		}
		return null; //never happens since #handleException() ALWAYS throws one
    }
    
    /**
     * includes comments too!
     * 
     * gives back a design regardless of whether its in the catalogue etc. this
     * method is used to show the designs details. hence it must be able to show designs which are not
     * in the catalogue or are not approved.
     * 
     * @see  ApplicationLayerService#getDesign(Integer)
     */
    public Design getDesign(Integer designUid, HttpServletRequest request) throws ApplicationException {
    	
    	try {
    		uk.co.maxant.gwtdemo.server.po.Design design = designService.getDesign(designUid);
    		
    		Design d = mapAndBuildDesign(design);

    		//this method loads all comments too!
    		d.setComments(getComments(request, d.getUid()));
    		
            userService.incrementArtistViews(design.getParty().getUid());

    		return d;
		}catch(Exception e){
			handleException("ORCH014", "failed to get session info: " + e.getMessage(), e);
		}
		return null; //never happens since #handleException() ALWAYS throws one
    }

	/**
     * user must be in "backoffice" or "picking" role to call this method. 
     *
     * @param   filter  can be null, in which case its ignored.
     *
     * @return  DOCUMENT ME!
     *
     * @throws  ApplicationException  DOCUMENT ME!
     *
     * @see     ApplicationLayerService#getDesignsAdmin(String)
     */
    @RolesAllowed({Roles.BACKOFFICE, Roles.PICKING})
    public List<Design> getDesigns(String filter) throws ApplicationException {

        try {
            List<uk.co.maxant.gwtdemo.server.po.Design> designs = designService.getDesigns(filter);
            List<Design> ds = new ArrayList<Design>();

            for (uk.co.maxant.gwtdemo.server.po.Design design : designs) {
                Design d = mapDesign(design);
                ds.add(d);
                d.setArtist(mapArtist(design.getParty()));
    			d.setNumberOfComments(design.getDesignComments().size());

                userService.incrementArtistViews(design.getParty().getUid());
            }

            return ds;
		}catch(Exception e){
			handleException("ORCH015", "failed to get session info: " + e.getMessage(), e);
		}
		return null; //never happens since #handleException() ALWAYS throws one
    }

    /**
     * user must be in "backoffice" role to call this method. 
     * 
     * @return  loaded designs, ALL of them. artist are loaded.
     *
     * @throws  ApplicationException  DOCUMENT ME!
     *
     * @see     ApplicationLayerService#getDesignsAdminLight()
     */
    @RolesAllowed({Roles.BACKOFFICE, Roles.ARTIST})
    public List<Design> getAllDesigns(HttpServletRequest request) throws ApplicationException {
    	
    	try {
    		Party loggedInUser = null;
    		Artist lia = null;
    		List<uk.co.maxant.gwtdemo.server.po.Design> designs = null;
    		if(request.isUserInRole(Roles.BACKOFFICE)){
    			designs = designService.getDesigns(null);
    		}else if(request.isUserInRole(Roles.ARTIST)){
    			loggedInUser = userService.getLoggedinUser(request.getUserPrincipal());
    			lia = mapArtist(loggedInUser);
    			designs = designService.getDesignsForArtist(loggedInUser.getUid(), true);
    		}else{
    			throw new RuntimeException("code change required because of unexpected role");
    		}
    		List<Design> ds = new ArrayList<Design>();
    		
    		for (uk.co.maxant.gwtdemo.server.po.Design design : designs) {
    			Design d = mapDesign(design);
    			ds.add(d);
        		if(request.isUserInRole(Roles.BACKOFFICE)){
        			d.setArtist(mapArtist(design.getParty()));
        		}else if(request.isUserInRole(Roles.ARTIST)){
        			d.setArtist(lia);
        		}else{
        			throw new RuntimeException("code change required because of unexpected role");
        		}
    		}
    		
    		return ds;
		}catch(Exception e){
			handleException("ORCH016", "failed to get session info: " + e.getMessage(), e);
		}
		return null; //never happens since #handleException() ALWAYS throws one
    }
    
    /**
     * user must be in "backoffice" role to call this method. this method loads artists regardless.
     *
     * @param   filter  can be null, in which case its ignored.
     *
     * @return  DOCUMENT ME!
     *
     * @throws  ApplicationException  DOCUMENT ME!
     *
     * @see     ApplicationLayerService#getArtistsAdmin(String)
     */
    @RolesAllowed({Roles.BACKOFFICE, Roles.PICKING})
    public List<Artist> getArtists(String filter) throws ApplicationException {

        try {
            List<Party> artists = userService.getArtists(filter);
            List<Artist> as = new ArrayList<Artist>();

            for (Party user : artists) {
                Artist a = mapArtist(user);
                a.setDesigns(new ArrayList<Design>());

                for (uk.co.maxant.gwtdemo.server.po.Design design : user.getDesigns()) {
                    a.getDesigns().add(mapDesign(design));
                }
                as.add(a);
            }

            return as;
		}catch(Exception e){
			handleException("ORCH017", "failed to get session info: " + e.getMessage(), e);
		}
		return null; //never happens since #handleException() ALWAYS throws one
    }

    /**
     * @see  ApplicationLayerService#search(String)
     */
    public SearchResults search(String searchExpression, HttpServletRequest request)
        throws ApplicationException {

        try {
            DesignService.SearchResult result = designService.search(searchExpression);

            List<Design> ds = new ArrayList<Design>();
            List<Artist> as = new ArrayList<Artist>();

            for (uk.co.maxant.gwtdemo.server.po.Design design : result.designs) {
            	if(ds.size() >= MAX_NUMBER_OF_DESIGNS_TO_RETURN) {
            		break;
            	}
    			Design d = mapDesign(design);
    			d.setArtist(mapArtist(design.getParty()));
    			ds.add(d);
            }
            
            for (Party artist : result.artists) {
            	if(as.size() >= MAX_NUMBER_OF_ARTISTS_TO_RETURN) {
            		break;
            	}
                Artist a = mapArtist(artist);
                as.add(a);
            }
            
            return new SearchResults(ds, as);
		}catch(Exception e){
			handleException("ORCH018", "failed to get session info: " + e.getMessage(), e);
		}
		return null; //never happens since #handleException() ALWAYS throws one
    }

    /**
     * user must be in "backoffice" role.
     *
     * @param   design             DOCUMENT ME!
     *
     * @throws  ApplicationException  DOCUMENT ME!
     *
     * @see     DesignService#saveOrUpdate(uk.co.maxant.gwtdemo.server.po.Design, List, List, List)
     */
    @RolesAllowed({Roles.BACKOFFICE, Roles.ARTIST})
    public Integer saveOrUpdateDesign(Design design, HttpServletRequest request)
        throws ApplicationException {

    	try {
    		boolean sendMail = false;
    		Party loggedInUser = null;
            uk.co.maxant.gwtdemo.server.po.Design d = mapDesign(design);
            if(request.isUserInRole(Roles.BACKOFFICE)){
            	d.setParty(mapArtist(design.getArtist()));
            }else if(request.isUserInRole(Roles.ARTIST)){
            	sendMail = true;
    	    	loggedInUser = userService.getLoggedinUser(request.getUserPrincipal());
    	    	
    	    	//if its an artist, then they can only create/edit their own designs
    	    	if(!d.getParty().getUid().equals(loggedInUser.getUid())){
	    			throw new SecurityException("logged in user is not backoffice and only artist - they " +
	    					"are trying to add a design " + design.getName() + " which doesnt belong to them");
    	    	}
				d.setParty(loggedInUser);
            }else{
            	throw new SecurityException("unexpected role");
            }

            Integer uid = designService.saveOrUpdate(d, design.getTags(), request);
            
            if(sendMail){
            	String msg = "Hi,<br><br>" + loggedInUser.getFirstname() + " " + loggedInUser.getLastname() +
            			" (" + loggedInUser.getScreenname() + " / " + loggedInUser.getUid() + ") " +
    					"has updated their design (" + uid + "). This update is now live, so please ensure you check " +
						"the content. If it is inappropriate you can update it using the admin screens.";
            	mailService.sendMail(
            			configService.value(ConfigService.emailOps), 
            			"[Design Update] Design " + uid,
            			msg);
            }
            
            return uid;
            
		}catch(Exception e){
			handleException("ORCH019", "failed to get session info: " + e.getMessage(), e);
		}
		return null; //never happens since #handleException() ALWAYS throws one
    }

    /**
     * user must be in "backoffice" role.
     *
     * @param   artist  DOCUMENT ME!
     *
     * @throws  ApplicationException  DOCUMENT ME!
     *
     * @see     PartyService#updateUser(Party, Boolean, boolean, boolean)
     */
    @RolesAllowed(Roles.BACKOFFICE)
    public Integer updateArtist(Artist artist) throws ApplicationException {

        try {
        	return userService.updateUser(mapArtist(artist), true, true, false);
		}catch(Exception e){
			handleException("ORCH020", "failed to get session info: " + e.getMessage(), e);
		}
		return null; //never happens since #handleException() ALWAYS throws one
    }
    
    /**
     * user must be in "registered" role.
     * 
     * here a user is updating their own artist profile.
     *
     * @see     PartyService#updateUser(Users, Boolean, boolean, boolean)
     */
    @RolesAllowed(Roles.REGISTERED)
    public Integer updateArtistProfile(Artist artist, HttpServletRequest request) throws ApplicationException {
    	
    	try {
    		//we can do this all here, since its only a mapping issue.

    		Party loggedInUser = userService.getLoggedinUser(request.getUserPrincipal());
            
            loggedInUser.setArtistDescription(artist.getDescription());

        	String msg = "Hi,<br><br>" + loggedInUser.getFirstname() + " " + loggedInUser.getLastname() +
        			" (" + loggedInUser.getScreenname() + " / " + loggedInUser.getUid() + ") " +
					"has updated their profile. This update is now live, so please ensure you check " +
					"the content. If it is inappropriate you can update it using the admin screens.";
        	mailService.sendMail(
        			configService.value(ConfigService.emailOps), 
        			"[Artist Update] Artist " + loggedInUser.getUid(),
        			msg);

            return loggedInUser.getUid();
            
		}catch(Exception e){
			handleException("ORCH021", "failed to get session info: " + e.getMessage(), e);
		}
		return null; //never happens since #handleException() ALWAYS throws one
    }
    
    /**
     * update their own account. if password is null, its not updated. if email is null its not updated.
     *
     * @param   artist  DOCUMENT ME!
     * 
     * @return true, if the user needs to login again. ie their password has changed.
     *
     * @throws  ApplicationException  DOCUMENT ME!
     *
     * @see     PartyService#updateUser(Users, Boolean, boolean, boolean)
     */
    @RolesAllowed(Roles.REGISTERED)
    public Integer updateCustomer(User customer, String newPassword, HttpServletRequest request) throws ApplicationException {
    	
    	try {
    		Party u = userService.getLoggedinUser(request.getUserPrincipal());
    		customer.setUid(u.getUid()); //forces logged in user to be edited

    		String email = customer.getEmail();
    		if(email == null){
    			//is the case if it hasnt changed and shouldnt be updated
    			email = u.getLogin();
    		}
    		
    		u = mapCustomer(customer);
    		if(newPassword != null){
    			u.setPassword(newPassword);
    		}
    		Integer uid = userService.updateUser(u, false, false, false);
    		
    		String siteName = configService.value(ConfigService.siteName);
            String subject = "Your account with " + siteName;
    		String message = "Your account with " + siteName + " has been updated.<br><br>" +
    				"If you did not change it, please contact us immediately.";
            mailService.sendMail(email, subject, message);
    		
            return uid;
		}catch(Exception e){
			handleException("ORCH022", "failed to get session info: " + e.getMessage(), e);
		}
		return null; //never happens since #handleException() ALWAYS throws one
    }

    /**
     * @see PartyService#cleanArtistViews()
     */
	public void cleanArtistViews() throws ApplicationException {
        try {
            userService.cleanArtistViews();
		}catch(Exception e){
			handleException("ORCH023", "failed to get session info: " + e.getMessage(), e);
		}
	}

	/**
	 * @see PartyService#getDesignViewStats()
	 */
	public List<Artist> getArtistViewCloud(HttpServletRequest request) throws ApplicationException {
		try {
			List<Artist> cloud = new ArrayList<Artist>();
			List<ArtistViewStat> stats = userService.getArtistViewStats();
			for(ArtistViewStat avs : stats){
				cloud.add(mapArtist(avs.artist));
			}
			
			return cloud;
		}catch(Exception e){
			handleException("ORCH024", "failed to get session info: " + e.getMessage(), e);
		}
		return null; //never happens since #handleException() ALWAYS throws one
	}

	/**
	 * gets all pending sales orders, which pickers need to have on their radar.
	 *
	 * @throws  ApplicationException  if anything goes wrong
	 *
	 * @see     SalesOrderService#getPendingSalesOrders()
	 */
	@RolesAllowed(Roles.PICKING)
	public List<SalesOrder> getPendingSalesOrders() throws ApplicationException {
		try {
			List<SalesOrder> sos = new ArrayList<SalesOrder>();
			for(uk.co.maxant.gwtdemo.server.po.SalesOrder salesOrder : salesOrderService.getPendingSalesOrders()){
				SalesOrder so = mapSalesOrderFully(salesOrder);
				sos.add(so);
			}
			return sos;
		}catch(Exception e){
			handleException("ORCH025", "failed to get session info: " + e.getMessage(), e);
		}
		return null; //never happens since #handleException() ALWAYS throws one
	}
	
	/**
	 * gets all discarded sales orders.
	 *
	 * @throws  ApplicationException  if anything goes wrong
	 *
	 * @see     SalesOrderService#getDiscardedSalesOrders()
	 */
	@RolesAllowed(Roles.PICKING)
	public List<SalesOrder> getDiscardedSalesOrders() throws ApplicationException {
		try {
			List<SalesOrder> sos = new ArrayList<SalesOrder>();
			for(uk.co.maxant.gwtdemo.server.po.SalesOrder salesOrder : salesOrderService.getDiscardedSalesOrders()){
				SalesOrder so = mapSalesOrderFully(salesOrder);
				sos.add(so);
			}
			return sos;
		}catch(Exception e){
			handleException("ORCH026", "failed to get session info: " + e.getMessage(), e);
		}
		return null; //never happens since #handleException() ALWAYS throws one
	}
	
	/**
	 * gets all problem sales orders, which pickers need to have on their radar.
	 *
	 * @throws  ApplicationException  if anything goes wrong
	 *
	 * @see     SalesOrderService#getPendingSalesOrders()
	 */
	@RolesAllowed(Roles.PICKING)
	public List<SalesOrder> getProblemSalesOrders() throws ApplicationException {
		try {
			List<SalesOrder> sos = new ArrayList<SalesOrder>();
			for(uk.co.maxant.gwtdemo.server.po.SalesOrder salesOrder : salesOrderService.getProblemSalesOrders()){
				SalesOrder so = mapSalesOrderFully(salesOrder);
				sos.add(so);
			}
			return sos;
		}catch(Exception e){
			handleException("ORCH027", "failed to get session info: " + e.getMessage(), e);
		}
		return null; //never happens since #handleException() ALWAYS throws one
	}

	/**
	 * gets all sales orders, which match the search string.
	 *
	 * @throws  ApplicationException  if anything goes wrong
	 *
	 * @see     SalesOrderService#searchSalesOrders()
	 */
	@RolesAllowed(Roles.PICKING)
	public List<SalesOrder> searchSalesOrders(String search) throws ApplicationException {
		try {
			List<SalesOrder> sos = new ArrayList<SalesOrder>();
			for(uk.co.maxant.gwtdemo.server.po.SalesOrder salesOrder : salesOrderService.searchSalesOrders(search)){
				SalesOrder so = mapSalesOrderFully(salesOrder);
				sos.add(so);
			}
			return sos;
		}catch(Exception e){
			handleException("ORCH028", "failed to get session info: " + e.getMessage(), e);
		}
		return null; //never happens since #handleException() ALWAYS throws one
	}
	
	/**
	 * gets all sales orders, which belong to the given customer.
	 *
	 * @throws  ApplicationException  if anything goes wrong
	 *
	 * @see     SalesOrderService#getCustomerSalesOrders()
	 */
	public List<SalesOrder> getCustomerSalesOrders(HttpServletRequest request) throws ApplicationException {

		if(request.getUserPrincipal() == null){
			//force a login
			throw new SecurityException("To get customer orders, the user must be authenticated.");
		}
		
		Party user = userService.getLoggedinUser(request.getUserPrincipal()); //cannot be null, since we checked that above...

		try {
			List<SalesOrder> sos = new ArrayList<SalesOrder>();
			for(uk.co.maxant.gwtdemo.server.po.SalesOrder salesOrder : 
								salesOrderService.getCustomerSalesOrders(user)){
				SalesOrder so = mapSalesOrderFully(salesOrder);
				sos.add(so);
			}
			return sos;
		}catch(Exception e){
			handleException("ORCH029", "failed to get session info: " + e.getMessage(), e);
		}
		return null; //never happens since #handleException() ALWAYS throws one
	}
	
	/**
	 * gets the given order.
	 * 
	 * checks security - either a picker should call this method, or the sales order owner.
	 *
	 * @param   artist  the artist containing the data to save.
	 *
	 * @throws  ApplicationException  if anything goes wrong
	 *
	 * @return null if not found or id is empty. otherwise, the fully loaded SO.
	 * 
	 * @see SalesOrderService#getSalesOrder(Integer)
	 */
	public SalesOrder getSalesOrder(String id, HttpServletRequest request) throws ApplicationException {
		
		try{
			if(id == null || id.trim().length() == 0){
				return null;
			}
			
			Integer uid = null;
			try{
				uid = Integer.parseInt(id.trim());
			}catch(NumberFormatException nfe){
				throw new ApplicationException("ORCH030", "'" + uid + "' is not an integer", "Please enter a valid number.");
			}
			
			boolean doit = false;
			uk.co.maxant.gwtdemo.server.po.SalesOrder so = null;
			if(request.isUserInRole(Roles.PICKING)){
				doit = true;
				so = salesOrderService.getSalesOrder(uid);
			}else{
				if(request.getUserPrincipal() != null){
					//perhaps the order belongs to them? if so, then continue.
					so = salesOrderService.getSalesOrder(uid);
					if(so.getPartyByCustomerUid().getLogin().equals(request.getUserPrincipal().getName())){
						doit = true;
					}
				}
			}
			
			if(doit){
				SalesOrder salesOrder = null;
				if(so != null){
					salesOrder = mapSalesOrderFully(so);
				}
				return  salesOrder;
			}else{
				throw new ApplicationException("ORCH031", "no rights to open order " + uid, "You do not have authorization to view sales order " + uid + ". Please log in again with sufficient rights and try again.");
			}
		}catch(Exception e){
			handleException("ORCH032", "failed to get session info: " + e.getMessage(), e);
		}
		return null; //never happens since #handleException() ALWAYS throws one
	}

    /**
     * @param customer - the customer to save
     * @param alsoArtist - they should be marked as also being an artist.
     * @throws  ApplicationException  if anything goes arye
     *
     * @see     PartyService#registerUser(Users)
     */
	public Integer registerCustomerOrArtist(User user, String password, boolean isArtist, 
			String artistsDescription, HttpServletRequest request) throws ApplicationException {
        try {
        	Party u = mapUser(
        			user, 
        			isArtist, 
        			artistsDescription, 
        			null /*since person has just registered, we definitely have no notes about them*/, 
        			false,
        			null /*a newly registered user is not staff, until they are upgraded to that...*/,
        			null /*not staff right now!*/,
        			null /*since person has just registered, we definitely have no notes about them*/
    			);

        	u.setArtistDescription(artistsDescription);
        	u.setPassword(password);
            u = userService.registerUser(u);

            try {
				sendRegistrationMail(u, request);
			} catch (UnsupportedEncodingException e) {
				throw new ApplicationException("ORCH033", "failed to send reg mail due to: " + e.getMessage(), "Could not send email confirming registration.");
			}
            
            return u.getUid();
		}catch(Exception e){
			handleException("ORCH034", "failed to get session info: " + e.getMessage(), e);
		}
		return null; //never happens since #handleException() ALWAYS throws one
    }

	private void sendRegistrationConfirmationMail(String email) throws ApplicationException {
		String siteName = configService.value(ConfigService.siteName);
		String url = configService.value(ConfigService.httpUrl);
        String subject = "Registration confirmed with " + siteName;
		String message = "Your registration with " + siteName + " is now confirmed.<br><br>" +
				"Edit your profile by logging in at " + url;
        mailService.sendMail(email, subject, message);
	}

	/**
	 * designed to resend this persons reg code to them.
	 */
	public void sendRegistrationMail(String email, HttpServletRequest request) throws ApplicationException {
		try{
			Party  u = userService.getUser(email);
			if(u == null){
				throw new ApplicationException("ORCH035", "failed to find user " + email, "Could not find email address.");
			}
			sendRegistrationMail(u, request);
		}catch(Exception e){
			handleException("ORCH036", "failed to get session info: " + e.getMessage(), e);
		}
	}

	private void sendRegistrationMail(Party u, HttpServletRequest request) throws ApplicationException, UnsupportedEncodingException {
		String siteName = configService.value(ConfigService.siteName);
		String url = configService.value(ConfigService.httpUrl);
		if(configService.boolValue(ConfigService.registrationForceConfirm)){
	        String subject = "Registered with " + siteName;
	        String url2 = url + "/confirmRegistration.jsp";
			String message = 
				"Thank you for registering with " + siteName + ".<br>" +
				"<br>" +
				"Please confirm your account now, by either clicking the following link or going to this page:<br>" +
				"<br>" +
				"&nbsp;&nbsp;&nbsp;&nbsp; <a href='" + 
					url2 + 
					"?email=" + 
					URLEncoder.encode(u.getLogin(), "UTF-8") + 
					"&regCode=" + 
					URLEncoder.encode(u.getRegCode(), "UTF-8") + 
					"'>" + url2 + "</a><br>" +
				"<br>" +
				"and entering your email address and the following registration code:<br>" +
				"<br>" +
				"&nbsp;&nbsp;&nbsp;&nbsp; " + u.getRegCode() + "<br>" +
				"<br>" +
				"some other text? the business to specify...";
	        mailService.sendMail(u.getLogin(), subject, message);
		}else{
	        String subject = "Registered with " + siteName;
			String message = 
				"Thank you for registering with " + siteName + ".<br>" +
				"<br>" +
				"Edit your profile by logging in at " + url + "<br>" +
				"<br>" +
				"some other text, for the business to specify...";
	        mailService.sendMail(u.getLogin(), subject, message);
		}
	}

	/**
	 * confirms the users registration.
	 */
	public void confirmUserRegistration(Party u, HttpServletRequest request) throws ApplicationException {
		try{
			userService.confirmUserRegistration(u);
			
            //now mail them
            sendRegistrationConfirmationMail(u.getLogin());
		}catch(Exception e){
			handleException("ORCH037", "failed to get session info: " + e.getMessage(), e);
		}
	}
	
	/**
	 * @see     DesignService#registerUser(Users)
	 */
	public void rateDesign(Integer designUid, HttpServletRequest request, int ratingGiven) throws ApplicationException {
		try {
			Party user = userService.getLoggedinUser(request.getUserPrincipal());
			designService.rateDesign(designUid, (user==null?null:user.getUid()), ratingGiven);
		}catch(Exception e){
			handleException("ORCH038", "failed to get session info: " + e.getMessage(), e);
		}
	}
	
	/**
	 * @see     SalesOrderService#updateOrderStatus(Users, OrderState, int, String)
	 */
    @RolesAllowed(Roles.PICKING)
	public void updateOrderStatus(HttpServletRequest request, OrderState requiredState, int soUid, String notes) throws ApplicationException {
		try {
			Party staff = userService.getLoggedinUser(request.getUserPrincipal());
			salesOrderService.updateOrderStatus(staff, requiredState, soUid, notes);
		}catch(Exception e){
			handleException("ORCH039", "failed to get session info: " + e.getMessage(), e);
		}
	}

	/**
	 * must be in picking role to do this.
	 * 
	 * @see     SalesOrderService#updatePaymentStatus(Users, OrderState, int, String, String)
	 */
    @RolesAllowed(Roles.PICKING)
	public void updatePaymentStatus(HttpServletRequest request, OrderState requiredState, int soUid, String notes, String txnId) throws ApplicationException {
		try {
			Party staff = userService.getLoggedinUser(request.getUserPrincipal());
			
			salesOrderService.updatePaymentStatus(staff, requiredState, soUid, notes, txnId);
		}catch(Exception e){
			handleException("ORCH040", "failed to get session info: " + e.getMessage(), e);
		}
	}
    
    /**
     * must be in picking role to do this.
     * 
     * @see     SalesOrderService#addShopNotes(Users, int, String)
     */
    @RolesAllowed(Roles.PICKING)
	public void addShopNotes(HttpServletRequest request, int soUid, String notes) throws ApplicationException {
    	try {
    		Party staff = userService.getLoggedinUser(request.getUserPrincipal());
    		salesOrderService.addShopNotes(staff, soUid, notes);
		}catch(Exception e){
			handleException("ORCH041", "failed to get session info: " + e.getMessage(), e);
		}
    }

    /**
     * must be in picking role to do this.
     * 
     * @see     SalesOrderService#markOrderAsNotNeedingAttention(Users, int, String)
     */
    @RolesAllowed(Roles.PICKING)
	public void markOrderAsNotNeedingAttention(HttpServletRequest request, int soUid, String notes) throws ApplicationException {
    	try {
    		Party staff = userService.getLoggedinUser(request.getUserPrincipal());
    		salesOrderService.markOrderAsNotNeedingAttention(staff, soUid, notes);
		}catch(Exception e){
			handleException("ORCH042", "failed to get session info: " + e.getMessage(), e);
		}
    }

    /**
     * must be in picking role to do this.
     * 
     * @see     SalesOrderService#addShopNotes(Users, int, String)
     */
    @RolesAllowed(Roles.PICKING)
	public void markOrderAsNeedingAttention(HttpServletRequest request, Integer orderNum, String msg) throws ApplicationException {
    	try {
    		Party staff = userService.getLoggedinUser(request.getUserPrincipal());
    		salesOrderService.markOrderAsNeedingAttention(staff, orderNum, msg);
		}catch(Exception e){
			handleException("ORCH043", "failed to get session info: " + e.getMessage(), e);
		}
    }
    
    /**
     * @see     DesignService#getAllComments(Integer)
     */
    @RolesAllowed(Roles.DESIGN_COMMENTS_MODERATOR)
    public List<DesignComment> getAllComments(HttpServletRequest request, Integer designUid) throws ApplicationException {
    	try {
    		List<uk.co.maxant.gwtdemo.server.po.DesignComment> comments = 
    						designService.getAllComments(designUid);
    		
    		List<DesignComment> dcs = new ArrayList<DesignComment>();
    		for(uk.co.maxant.gwtdemo.server.po.DesignComment comment : comments){
    			DesignComment dc = mapDesignComment(comment);
    			
    			dc.setCreator(mapUser(comment.getPartyByUserUid()));
    			dc.setDesign(mapDesign(comment.getDesign()));
    			dc.setModerator(mapStaff(comment.getPartyByModeratedByUid()));
    			
    			dcs.add(dc);
    		}
    		
    		return dcs;
		}catch(Exception e){
			handleException("ORCH044", "failed to get session info: " + e.getMessage(), e);
		}
		return null; //never happens since #handleException() ALWAYS throws one
    }
    
    /**
     * @see     DesignService#getComments(Integer)
     */
    public List<DesignComment> getComments(HttpServletRequest request, Integer designUid) throws ApplicationException {
    	try {
    		List<uk.co.maxant.gwtdemo.server.po.DesignComment> comments = 
				designService.getComments(designUid);

    		List<DesignComment> dcs = new ArrayList<DesignComment>();
    		for(uk.co.maxant.gwtdemo.server.po.DesignComment comment : comments){
    			DesignComment dc = mapDesignComment(comment);
    			
    			dc.setCreator(mapUser(comment.getPartyByUserUid()));
    			dc.setDesign(mapDesign(comment.getDesign()));
    			dc.setModerator(mapStaff(comment.getPartyByModeratedByUid()));
    			
    			dcs.add(dc);
    		}
    		
    		return dcs;
		}catch(Exception e){
			handleException("ORCH045", "failed to get session info: " + e.getMessage(), e);
		}
		return null; //never happens since #handleException() ALWAYS throws one
    }
    
    /**
     * must be registered
     * 
     * @see     DesignService#saveOrUpdate(uk.co.maxant.gwtdemo.server.po.DesignComment)
     */
    @RolesAllowed(Roles.REGISTERED)
    public Integer addDesignComment(HttpServletRequest request, DesignComment dc) throws ApplicationException {
    	try {
    		Party user = userService.getLoggedinUser(request.getUserPrincipal());
    		
    		uk.co.maxant.gwtdemo.server.po.DesignComment designComment = mapDesignComment(dc);
    		
    		designComment.setAdded(new Date());
    		designComment.setApproved(null);
    		designComment.setDesign(mapDesign(dc.getDesign()));
    		designComment.setRejected(null);
    		designComment.setUid(null);
    		designComment.setPartyByModeratedByUid(null);
    		designComment.setPartyByUserUid(user);
    		
    		Integer uid = designService.saveOrUpdate(designComment);

    		if(configService.boolValue(ConfigService.designCommentModerationRequired)){
    			if(configService.boolValue(ConfigService.designCommentModerationSendEmailToModerator)){
    				//notify by email!
    				String msg = "Design " + dc.getDesign().getUid() + " has had a comment added which " +
    						"requires moderation before it will be seen by other customers. Please " +
    						"approve, modify or reject this comment by logging in and viewing the " +
    						"designs details (click the link at the top of the viewer in the shop).";
    				mailService.sendMail(configService.value(ConfigService.emailOps),
    						"[Design Comment Moderation] Design " + dc.getDesign().getUid(), msg);
    			}
    		}
    		return uid;
		}catch(Exception e){
			handleException("ORCH046", "failed to get session info: " + e.getMessage(), e);
		}
		return null; //never happens since #handleException() ALWAYS throws one
    }
    
    /**
     * @see     DesignService#saveOrUpdate(uk.co.maxant.gwtdemo.server.po.DesignComment)
     */
    @RolesAllowed(Roles.DESIGN_COMMENTS_MODERATOR)
    public void approveDesignComment(HttpServletRequest request, DesignComment dc) throws ApplicationException {
    	try {
    		Party staff = userService.getLoggedinUser(request.getUserPrincipal());
    		
    		uk.co.maxant.gwtdemo.server.po.DesignComment designComment = mapDesignComment(dc);
    		
    		designComment.setApproved(new Date());
    		designComment.setRejected(null);
    		designComment.setPartyByModeratedByUid(staff);
    		
    		designService.saveOrUpdate(designComment);
		}catch(Exception e){
			handleException("ORCH047", "failed to get session info: " + e.getMessage(), e);
		}
    }
    
    /**
     * @see     DesignService#saveOrUpdate(uk.co.maxant.gwtdemo.server.po.DesignComment)
     */
    @RolesAllowed(Roles.DESIGN_COMMENTS_MODERATOR)
    public void rejectDesignComment(HttpServletRequest request, DesignComment dc) throws ApplicationException {
    	try {
    		Party staff = userService.getLoggedinUser(request.getUserPrincipal());
    		
    		uk.co.maxant.gwtdemo.server.po.DesignComment designComment = mapDesignComment(dc);
    		
    		designComment.setApproved(null);
    		designComment.setRejected(new Date());
    		designComment.setPartyByModeratedByUid(staff);
    		
    		designService.saveOrUpdate(designComment);
		}catch(Exception e){
			handleException("ORCH048", "failed to get session info: " + e.getMessage(), e);
		}
    }

    /**
     * must be a moderator to call this!
     * 
     * @see     DesignService#getDesignsPendingModeration()
     */
    @RolesAllowed(Roles.DESIGN_COMMENTS_MODERATOR)
    public List<Design> getDesignsPendingModeration(HttpServletRequest request) throws ApplicationException {
    	try {
    		List<uk.co.maxant.gwtdemo.server.po.Design> designs = designService.getDesignsPendingModeration();
    		return mapAndBuildDesigns(designs);
		}catch(Exception e){
			handleException("ORCH049", "failed to get session info: " + e.getMessage(), e);
		}
		return null; //never happens since #handleException() ALWAYS throws one
    }

    /**
     * creates a list of design TOs based upon the given POs. 
     * 
     * sets the artist.
     * 
     * sets the number of design comments, although this number includes non-moderated or 
     * rejected ones, for optimisation reasons.
     * @throws ApplicationException 
     */
	private List<Design> mapAndBuildDesigns(
			List<uk.co.maxant.gwtdemo.server.po.Design> designs) {
		
		List<Design> ds = new ArrayList<Design>();
		for (uk.co.maxant.gwtdemo.server.po.Design design : designs) {

			Design d = mapAndBuildDesign(design);
			if(d != null){
				ds.add(d);
			}
		}
		
		return ds;
	}

	/**
	 * @see #mapAndBuildDesigns(List)
	 * 
	 * @return a mapped design
	 */
	private Design mapAndBuildDesign(uk.co.maxant.gwtdemo.server.po.Design design) {
		
		Design d = mapDesign(design);
		
		d.setArtist(mapArtist(design.getParty()));
		d.setNumberOfComments(design.getDesignComments().size());

		return d;
	}

	/**
	 * even if it fails, it does so quietly, to discourage hackers :-)
	 */
	public void remindPassword(HttpServletRequest request, String email) {
		try{
			Party user = userService.getUser(email);
			String siteName = configService.value(ConfigService.siteName);
			String msg = 
					"Hi " + user.getFirstname() + ",<br>" +
					"<br>" +
					"You have requested your password for " + siteName  + 
					" (" + configService.value(ConfigService.httpUrl) + ")." +
					"<br><br>Your password is: " + user.getPassword() +
					"<br><br>You should now be able to log into our site. If you did not request this email, please " +
					"inform us immediately and we will look into it.<br>" +
					"<br>" +
					"Your sincerely,<br>" + siteName;
			
			mailService.sendMail(email, "Password for " + siteName, msg);
		}catch(Exception e){
			//:-( but who really cares. lets log it...
			log.warn("failed to send user " + email + " their password", e);
		}
	}

	/**
	 * designed to be called by a user to get their own profile.
	 * if they are not an artist, then their description is simply null.
	 * 
	 * @see PartyService#getLoggedinUser(Principal)
	 */
	@RolesAllowed(Roles.REGISTERED)
	public Artist getLoggedInArtist(HttpServletRequest request) throws ApplicationException {
		try{
			Party user = userService.getLoggedinUser(request.getUserPrincipal());
			Artist a = mapArtist(user);
			return a;
		}catch(Exception e){
			handleException("ORCH050", "failed to get session info: " + e.getMessage(), e);
		}
		return null; //never happens since #handleException() ALWAYS throws one
	}
	
	@Override
	public Artist mapArtist(Party u){
		
		Artist a = super.mapArtist(u);

		//does their image exist?
		if(path != null && a != null){
			File f = new File(path, "images/artists/artist_" + a.getUid() + ".jpg");
			a.setImageExists(f.exists());
		}

        return a;
	}
	
	@Override
	public Design mapDesign(uk.co.maxant.gwtdemo.server.po.Design design){
		
		Design d = super.mapDesign(design);
		
		//does their image exist?
		if(path != null && d != null){
			File f = new File(path, "images/designs/design_" + d.getUid() + ".jpg");
			d.setImageExists(f.exists());
		}
		
		return d;
	}
	
	/**
	 * TODO add ability to call this!
	 */
	@RolesAllowed(Roles.BACKOFFICE)
	public void banUser(User user){
		userService.banUser(user.getUid());
	}
	
	/**
	 * TODO add ability to call this!
	 */
	@RolesAllowed(Roles.BACKOFFICE)
	public void unbanUser(User user){
		userService.unbanUser(user.getUid());
	}

	/**
	 * provides a string containing transaction stats. as well as refreshes those stats. see
	 * hibernate for more info.
	 */
	public String logDbStats(){
		Statistics stats = this.hibernateTemplate.getSessionFactory().getStatistics();

		long numTx = stats.getTransactionCount() - lastStats.numTx;
		long olFailures = stats.getOptimisticFailureCount() - lastStats.olFailures;
		long qs = stats.getQueries().length - lastStats.queries.length;
		
		String qss = "";
		if(stats != null){
			for(String s : stats.getQueries()){
				qss += s + "<hr>";
			}
		}else{
			qss = "null";
		}
		qss += "&nbsp;"; //just in case empty
		
		String s = 
				"Stats, compared to last call:<br>" +
				"<table border='1'>" +
				"<tr>" +
				"	<td align='right'>Number of transactions</td>" +
				"	<td>" + numTx + "</td>" +
				"</tr>" +
				"<tr>" +
				"	<td align='right'>Optimistic Lock Failures</td>" +
				"	<td>" + olFailures + "</td>" +
				"</tr>" +
				"<tr>" +
				"	<td align='right'>Queries</td>" +
				"	<td>" + qs + "</td>" +
				"</tr>" +
				"<tr>" +
				"	<td align='right'>All</td>" +
				"	<td>" + stats + "</td>" +
				"</tr>" +
				"<tr>" +
				"	<td align='right'>Old</td>" +
				"	<td>" + lastStats.orig + "</td>" +
				"</tr>" +
				"<tr>" +
				"	<td align='right'>Queries</td>" +
				"	<td>" + qss + "</td>" +
				"</tr>" +
				"</table>";

		this.lastStats = new Stats(stats);

		stats.clear();
		
		return s;
	}
	
	private static class Stats {
		long numTx;
		
		String[] queries;
		
		long olFailures;
		
		String orig;
		
		public Stats(){
			queries = new String[0];
		}
		
		public Stats(Statistics newStats){
			this.numTx = newStats.getTransactionCount();
			this.queries = newStats.getQueries();
			this.olFailures = newStats.getOptimisticFailureCount();
			this.orig = newStats.toString();
		}
		
		
	}
	
}
