/*  
 * Copyright (c) 2009 Ant Kutschera, maxant
 * 
 * This file is part of the maxant GWT Demo.
 * 
 * maxant GWT Demo is free software: you can redistribute it and/or modify
 * it under the terms of the Lesser GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * maxant GWT Demo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Lesser GNU General Public License for more details.
 * You should have received a copy of the Lesser GNU General Public License
 * along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */

package uk.co.maxant.gwtdemo.server;

import java.security.Principal;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.hibernate.HibernateException;
import org.hibernate.SQLQuery;
import org.hibernate.Session;
import org.hibernate.SessionFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.orm.hibernate3.HibernateCallback;
import org.springframework.orm.hibernate3.HibernateTemplate;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Isolation;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;

import uk.co.maxant.gwtdemo.client.ApplicationException;
import uk.co.maxant.gwtdemo.client.Roles;
import uk.co.maxant.gwtdemo.server.po.ArtistView;
import uk.co.maxant.gwtdemo.server.po.Party;
import uk.co.maxant.gwtdemo.server.po.Role;
import uk.co.maxant.gwtdemo.server.util.NotesFormatter;
import uk.co.maxant.gwtdemo.server.util.PasswordGenerator;
import uk.co.maxant.gwtdemo.server.util.ServerFormats;

/**
 * used to manage users.
 */
@Service("partyService")
@Transactional(propagation=Propagation.REQUIRED, isolation=Isolation.DEFAULT, rollbackFor=ApplicationException.class)
public class PartyService {

	protected static final long VIEW_STATS_PERIOD = ServerFormats._28_DAYS;

	protected static final long VIEW_STATS_RELOAD_PERIOD = 6 * ServerFormats.HOUR;

	private List<ArtistViewStat> artistViewStatCache = null; //forces cache reload first time
	private int artistViewStatCacheNumberReads = 0;
	private Date artistViewStatCacheLastReload = new Date(); //cant be null since it causes a NPE

	private HibernateTemplate hibernateTemplate;

	@Autowired
	public void setSessionFactory(SessionFactory sessionFactory) {
		this.hibernateTemplate = new HibernateTemplate(sessionFactory);
	}

	@Autowired
	protected ConfigService configService;

	public PartyService(){};
	
	/**
	 * get the currently logged in user.
	 * @param request servlet request containing the security principal.
	 * @return null if not logged in, otherwise the user.
	 * @throws ApplicationException should anything go wrong
	 */
	@SuppressWarnings("unchecked")
	public Party getLoggedinUser(Principal principal) throws ApplicationException {
		if(principal == null){
			return null;
		}
			
		List<Party> users = 
			(List<Party>)this.hibernateTemplate.find(
					"from Party u where u.login = ?", 
					principal.getName().toLowerCase().trim());
		if(users.size() != 1){
			throw new ApplicationException("USES001", "no unique user found", "Data error. No unique user found.");
		}
		return users.get(0);
	}

	/**
	 * @param u the user to query
	 * @return a list of roles. never null.
	 * @throws ApplicationException if anything goes wrong 
	 */
	@SuppressWarnings("unchecked")
	public List<String> getRoles(Party u) throws ApplicationException {
		List<String> roles = new ArrayList<String>();
		
		List<Role> rs = this.hibernateTemplate.find("from Role r where r.login = ?", u.getLogin().toLowerCase().trim());
		for(Role r : rs){
			roles.add(r.getRole());
		}

		return roles;
	}

	/**
	 * get fully loaded artists matching the filter.
	 * @param filter filter to compare against.
	 * @return a list of fully loaded artists to be displayed eg in artist admin
	 * @throws ApplicationException in case of errors
	 */
	@SuppressWarnings("unchecked")
	public List<Party> getArtists(String filter) throws ApplicationException {
		List<Party> artists = new ArrayList<Party>();

		//TODO really could do with optimizing this with a stored proc, since we dont want to load ALL artists into memory! or use with paging...
		List<Party> us = this.hibernateTemplate.find("from Party u where u.isArtist = true");

		if(filter != null){
			filter = filter.trim().toLowerCase();
			filter = ".*" + filter + ".*";
		}
		
		for(Party u : us){
				
			boolean matches = false;
			if(filter == null){
				matches = true;
			}else {
				List<String> stringsToCompare = new ArrayList<String>();
				
				stringsToCompare.add(String.valueOf(u.getUid()));
				stringsToCompare.add(String.valueOf(u.getFirstname().toLowerCase()));
				stringsToCompare.add(String.valueOf(u.getLastname().toLowerCase()));
				stringsToCompare.add(String.valueOf(u.getLogin().toLowerCase()));
				if(u.getScreenname() != null){
					stringsToCompare.add(String.valueOf(u.getScreenname().toLowerCase()));
				}

				for(String s : stringsToCompare){
					if(s.toLowerCase().matches(filter)){
						matches = true;
						break;
					}
				}
			}
			if(matches){
				artists.add(u);
			}
		}
		
		return artists;
	}
	
	/**
	 * get all shallow loaded artists.
	 * @return a list of shallow loaded artists to be displayed eg in design admin
	 * @throws ApplicationException in case of errors
	 */
	@SuppressWarnings("unchecked")
	public List<Party> getAllArtists() throws ApplicationException {
		List<Party> us = this.hibernateTemplate.find("from Party u where u.isArtist = true");
		
		for(Party u : us){
			u.getFirstname();
		}
		return us;
	}
	
	/**
	 * updates the user object. 
	 * 
	 * @param artistPerspective if the context is that we are updating an artist then set this to true. if we are just updating a customer, then we cannot know eg artistDescription since its missed out of the mapping. so leave it as it is.
	 * @return the uid of the saved object
	 */
	public Integer updateUser(Party u, Boolean isBackofficeCallingThisMethod, boolean artistPerspective, boolean staffPerspective) throws ApplicationException {
		
		Party user = null;
		if(u.getUid() == null){
			throw new IllegalArgumentException("missing user UID. if you want to save a user, call a register method instead.");
		}else{
			user = (Party)this.hibernateTemplate.load(Party.class, u.getUid());
			String originalEmail = user.getLogin().toLowerCase();
			
			if(isBackofficeCallingThisMethod){
				user.setCustomerNotes(u.getCustomerNotes());

				//only backoffice can decide if its staff or not
				if(staffPerspective){
					user.setIsStaff(u.isIsStaff());
					if(user.isIsStaff()){
						user.setStaffNotes(u.getStaffNotes());
					}else{
						//dont delete - might be useful info later on?
					}
				}else{
					//we simply dont know, so leave it as it was
				}

				//backoffice an make you an artist as well as remove you as an artist
				if(artistPerspective){
					user.setIsArtist(u.isIsArtist());
					if(user.isIsArtist()){
						user.setArtistDescription(u.getArtistDescription());
						user.setArtistNotes(u.getArtistNotes());
					}else{
						//dont delete - might be useful info later on?
					}
				}else{
					//we simply dont know, so leave it as it was
				}
			}else{
				if(artistPerspective){
					if(!user.isIsArtist() && u.isIsArtist()){
						//they want to become an artist
						user.setIsArtist(true);
						user.setArtistDescription(u.getArtistDescription());
					}
				}else{
					//we simply dont know, so leave it as it was
				}
			}

			user.setFirstname(u.getFirstname());
			user.setLastname(u.getLastname());
			user.setJoinedMailingList(u.isJoinedMailingList());
			user.setMobile(u.getMobile());
			user.setPhone(u.getPhone());
			user.setTitle(u.getTitle());
			
			//email address and roles
			if(u.getLogin() != null){
				String login = u.getLogin().toLowerCase();

				//check for existing email address
				@SuppressWarnings("unchecked")
				List<Party> us = this.hibernateTemplate.find("from Party u where u.login = ?", login);
				if(!us.isEmpty()){
					throw new ApplicationException("USES006", "duplicate user login would occur if we saved this user", "The email address is already registered. Please ensure it is typed correctly and try again.");
				}

				user.setLogin(login);
				
				//need to update all roles too!
				@SuppressWarnings("unchecked")
				List<Role> roles = this.hibernateTemplate.find("from Role r where r.login = ?", originalEmail);
				for(Role r : roles){
					r.setLogin(login);
				}
			}
			
			//is isArtist, then ensure they have that role too
			if(user.isIsArtist()){
				boolean found = false;
				@SuppressWarnings("unchecked")
				List<Role> roles = this.hibernateTemplate.find("from Role r where r.login = ?", u.getLogin());
				for(Role r : roles){
					if(r.getRole().equals(Roles.ARTIST)){
						found = true;
						break;
					}
				}
				if(!found){
					Role r = new Role();
					r.setLogin(user.getLogin());
					r.setRole(Roles.ARTIST);
					this.hibernateTemplate.save(r);
				}
			}else{
				//ensure there are no artist role saved!
				@SuppressWarnings("unchecked")
				List<Role> roles = this.hibernateTemplate.find("from Role r where r.login = ?", u.getLogin());
				for(Role r : roles){
					if(r.getRole().equals(Roles.ARTIST)){
						this.hibernateTemplate.delete(r);
					}
				}
			}
			
			if(u.getPassword() != null){
				//their password has changed
				user.setPassword(u.getPassword());
			}

			if(!user.getScreenname().equals(u.getScreenname())){
				//they have changed it!
				//check for existing screen name
				@SuppressWarnings("unchecked")
				List<Party> us = this.hibernateTemplate.find("from Party u where u.screenname = ?", u.getScreenname());
				if(!us.isEmpty()){
					throw new ApplicationException("USES007", "duplicate user screenname would occur if we saved this user", "This screen name is already registered. Please try a different one.");
				}
				user.setScreenname(u.getScreenname());
			}

		}
		
		return user.getUid();
	}
	
	/**
	 * registers a customer/artist/staff (altho not marked as staff!). 
	 * as such this method just creates a base user.
	 * returns the saved object, including its URL.
	 */
	@SuppressWarnings("unchecked")
	public Party registerUser(Party u) throws ApplicationException {
		
		//check for existing email address
		List<Party> us = this.hibernateTemplate.find("from Party u where u.login = ?", u.getLogin().toLowerCase().trim());
		if(!us.isEmpty()){
			throw new ApplicationException("USES009", "duplicate user login would occur if we saved this user", "The email address is already registered. Please ensure it is typed correctly and try again, or click the link to get your password sent to you.");
		}

		//check for existing screen name
		us = this.hibernateTemplate.find("from Party u where u.screenname = ?", u.getScreenname());
		if(!us.isEmpty()){
			throw new ApplicationException("USES010", "duplicate user screenname would occur if we saved this user", "This screen name is already registered. Please try a different one.");
		}
		
		Party user = null;

		u.setLogin(u.getLogin().toLowerCase()); //ensure its lower case!
		u.setRegistrationDate(new Date());
		u.setIsBlocked(false);
		
		if(configService.boolValue(ConfigService.registrationForceConfirm)){
			u.setRegCode(PasswordGenerator.generatePassword());
		}
			
		Integer uid = (Integer)this.hibernateTemplate.save(u);
		user = u;
		user.setUid(uid);
		user.setRegCode(u.getRegCode()); //since caller may need it when sending a mail
		
		//now add roles!
		//add user as registered to start with.
		Role r = new Role(user.getLogin().toLowerCase(), Roles.REGISTERED);
		this.hibernateTemplate.save(r);
		
		return user;
	}
	
	/**
	 * confirms the Party registration.
	 */
	public void confirmUserRegistration(Party u) throws ApplicationException {
		Party user = (Party)this.hibernateTemplate.load(Party.class, u.getUid());
		user.setRegCode(null);
	}
	
	/**
	 * not public, since this should only be called in the context of the orchestration service, who isnt
	 * just going to give away anyones details!
	 * 
	 * gets the user if they exist.
	 * @return null if none found.
	 * @throws ApplicationException 
	 */
	@SuppressWarnings("unchecked")
	protected Party getUser(String emailAddress) throws ApplicationException {

		if(emailAddress == null){
			return null;
		}
		
		Party user = null;

		List<Party> us = this.hibernateTemplate.find("from Party u where u.login = ?", emailAddress.toLowerCase().trim());
		
		for(Party u : us){
			if(user == null){
				user = u;
			}else{
				throw new IllegalStateException("two Party have the same email address!");
			}
		}
		
		return user;
	}
	
	/**
	 * not public, since this should only be called in the context of the orchestration service, who isnt
	 * just going to create random users!
	 * 
	 * creates a new user (customer) for a new order.
	 * 
	 * @throws ApplicationException 
	 */
	protected Party createAnonymousUser(String email, String name) throws ApplicationException {
		Party user = new Party();
		
		NotesFormatter.appendCustomerNotes("Customer created by user service for their first order. User was not logged in and email address does not already exist.", user);

		if(email == null){
			email = "unknown@not_supplied.com"; //happens for discarded orders for example
		}
		
		String fn = name;
		if(fn == null) fn = ""; //happens for discarded orders for example
		String ln = "";
		int i = fn.indexOf(' ');
		if(i > -1){
			ln = fn.substring(i);
			fn = fn.substring(0, i);
		}
		user.setFirstname(fn);
		user.setLastname(ln);

		user.setIsArtist(false);
		user.setIsStaff(false);
		user.setJoinedMailingList(false); //since user hasnt consented, we set false as default.
		user.setJoinedMobileList(false); //since user hasnt consented, we set false as default.
		user.setLogin(email.toLowerCase().trim());
		user.setPassword(PasswordGenerator.generatePassword());
		user.setRegistrationDate(new Date());
		user.setIsBlocked(false);

		if(configService.boolValue(ConfigService.registrationForceConfirm)){
			user.setRegCode(PasswordGenerator.generatePassword());
		}
		
		Integer userUid = (Integer)this.hibernateTemplate.save(user);
		
		user.setUid(userUid);
		
		//now add roles!
		//add user as registered to start with.
		Role r = new Role(user.getLogin().toLowerCase(), Roles.REGISTERED);
		this.hibernateTemplate.save(r);
		
		return user;
	}
	

	public void banUser(Integer uid){
		Party user = (Party)this.hibernateTemplate.load(Party.class, uid);
		user.setIsBlocked(true);
	}
	
	public void unbanUser(Integer uid){
		Party user = (Party)this.hibernateTemplate.load(Party.class, uid);
		user.setIsBlocked(false);
	}

	/**
	 * gets design importance. gets top ten designs, by view frequency
	 * in the last period. if there are not ten designs, then it gets 
	 * random ones.
	 * 
	 * based on cached results. reloads cache if this method is called every 100 times
	 * or every 6 hours.
	 * 
	 * @see VIEW_STATS_PERIOD
	 * @return a sorted list of designs. no duplicates. no nulls.
	 * @throws ApplicationException if anything goes wrong
	 */
	public List<ArtistViewStat> getArtistViewStats() throws ApplicationException {
		artistViewStatCacheNumberReads++;
		boolean moreThanReloadPeriod = System.currentTimeMillis() - artistViewStatCacheLastReload.getTime() > VIEW_STATS_RELOAD_PERIOD;
		if(artistViewStatCache == null || //never been loaded before
				artistViewStatCacheNumberReads > 100 || //been read 100 times
				moreThanReloadPeriod){ //old

			//never been read, been read too many times or too old => reload stats!
			//so store cache and reset reload criteria
			artistViewStatCache = getStats();
			artistViewStatCacheNumberReads = 0;
			artistViewStatCacheLastReload = new Date();
		}			
		return artistViewStatCache;
	}
	
	@SuppressWarnings("unchecked")
	private List<ArtistViewStat> getStats() throws ApplicationException {
		List<Party> artists = getArtists(null/*get them all*/);
		int numberToGet = 10; //TODO configurable?
		Set<ArtistViewStat> stats = new HashSet<ArtistViewStat>();
		List<ArtistView> avs = this.hibernateTemplate.find("from ArtistView av where av.party.isArtist = true");
		for(ArtistView av : avs){

			//dont have equals etc in generated objects. so check each ID...
			ArtistViewStat stat = null;
			for(ArtistViewStat avstat : stats){
				if(avstat.artist.getUid().equals(av.getParty().getUid())){
					stat = avstat;
					break;
				}
			}
			if(stat == null){
				//new artist - create a stat, if artist is relevant
				for(Party a : artists){
					if(a.getUid().equals(av.getParty().getUid())){
						//its displayable, so add its stats
						stat = new ArtistViewStat(av.getParty());
						stats.add(stat);
						break;
					}
				}
			}
			if(stat != null){
				stat.numberOfViewsInPeriod += av.getNumberOfViews();
			}
		}
		
		//we have all the stats now - build a list of the top 10.
		List<ArtistViewStat> sortedStats = new ArrayList<ArtistViewStat>(stats);
		Collections.sort(sortedStats);

		//add the top ones now - and fill in with random others, should that not be enough
		stats.clear();
		int max = numberToGet;
		if(sortedStats.size() < max) max = sortedStats.size();
		for(int i = 0; i < max; i++){
			stats.add(sortedStats.get(i));
		}

		int i = 0;
		Collections.shuffle(artists); //randomize
		while(stats.size() < numberToGet && i < artists.size()){
			//add others, with low stats
			Party potentialToAdd = artists.get(i++);
			boolean found = false;
			for(ArtistViewStat avstat : stats){
				if(avstat.artist.getUid().equals(potentialToAdd.getUid())){
					found = true;
					break;
				}
			}
			if(found){
				continue; //start again
			}else{
				ArtistViewStat gvstat = new ArtistViewStat(potentialToAdd);
				stats.add(gvstat);
			}
		}

		sortedStats = new ArrayList<ArtistViewStat>(stats);
		Collections.sort(sortedStats); //resort and return sorted
		
		return sortedStats;
	}

	/**
	 * struct for containing how often each artist has been viewed
	 */
	protected static class ArtistViewStat implements Comparable<ArtistViewStat> {
		protected Party artist;
		protected int numberOfViewsInPeriod;
		
		public ArtistViewStat(Party artist){
			this.artist = artist;
			this.numberOfViewsInPeriod = 0;
		}
		
		public int compareTo(ArtistViewStat that) {
			return new Integer(that.numberOfViewsInPeriod).compareTo(new Integer(this.numberOfViewsInPeriod));
		}
	}

	/**
	 * increments the number of views for the given artist by one.
	 * 
	 * snychronized because there is a danger that two threads running at the same time could
	 * cause two artistView entries in the DB. our logic depends on only one being there.
	 * 
	 * @param artistUid
	 * @throws ApplicationException if anything goes wrong
	 */
	@SuppressWarnings("unchecked")
	public synchronized void incrementArtistViews(Integer artistUid) throws ApplicationException {
		List<ArtistView> avs = this.hibernateTemplate.find("from ArtistView av where av.party.uid = ? and av.day = now()", artistUid);
		if(avs.isEmpty()){
			Party artist = new Party();
			artist.setUid(artistUid);
			ArtistView av = new ArtistView(artist, new Date(), 1);
			this.hibernateTemplate.save(av);
		}else if(avs.size() == 1){
			ArtistView av = avs.get(0);
			av.setNumberOfViews(av.getNumberOfViews() + 1);
		}else{
			//strange... merge and delete... this is more of a work around than anything...
			//must be related to concurrency? TODO find out more...
			List<ArtistView> toDelete = new ArrayList<ArtistView>();
			ArtistView fv = avs.get(0);
			for(int i = 1; i < avs.size(); i++){
				ArtistView av = avs.get(i);
				fv.setNumberOfViews(fv.getNumberOfViews() + av.getNumberOfViews());
				toDelete.add(av);
			}
			
			for(ArtistView av : toDelete){
				this.hibernateTemplate.delete(av);
			}
		}
	}
	
	public void cleanArtistViews() throws ApplicationException {
		
		this.hibernateTemplate.execute(new HibernateCallback() {
			public Object doInHibernate(Session session)
					throws HibernateException, SQLException {
				
				Date date = new Date(System.currentTimeMillis() - VIEW_STATS_PERIOD);
				SQLQuery q = session.createSQLQuery("delete from artist_view where day < ?");
				q.setParameter(0, date);
				q.executeUpdate();
				return null;
			}
		});
	}

	
}
