/*  
 * Copyright (c) 2009 Ant Kutschera, maxant
 * 
 * This file is part of the maxant GWT Demo.
 * 
 * maxant GWT Demo is free software: you can redistribute it and/or modify
 * it under the terms of the Lesser GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * maxant GWT Demo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Lesser GNU General Public License for more details.
 * You should have received a copy of the Lesser GNU General Public License
 * along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */

package uk.co.maxant.gwtdemo.server;

import java.math.BigDecimal;
import java.math.MathContext;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;

import org.hibernate.SessionFactory;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.orm.hibernate3.HibernateTemplate;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Isolation;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;

import uk.co.maxant.gwtdemo.client.ApplicationException;
import uk.co.maxant.gwtdemo.client.to.OrderState;
import uk.co.maxant.gwtdemo.server.po.Party;
import uk.co.maxant.gwtdemo.server.po.SalesOrder;
import uk.co.maxant.gwtdemo.server.po.SalesOrderItem;
import uk.co.maxant.gwtdemo.server.util.BasketItem;
import uk.co.maxant.gwtdemo.server.util.LogHelper;
import uk.co.maxant.gwtdemo.server.util.NotesFormatter;
import uk.co.maxant.gwtdemo.server.util.PaymentDetails;
import uk.co.maxant.gwtdemo.server.util.SalesOrderHelper;
import uk.co.maxant.gwtdemo.server.util.ServerFormats;

/**
 * used to manage orders.
 */
@Service("salesOrderService")
@Transactional(
	propagation = Propagation.REQUIRED,
	isolation = Isolation.DEFAULT,
	rollbackFor = ApplicationException.class
)
public class SalesOrderService {

	private static Logger log = LogHelper.getLog();

	/** DOCUMENT ME! */
	private HibernateTemplate hibernateTemplate;

	/** DOCUMENT ME! */
	@Autowired
	private ConfigService configService;

	/** DOCUMENT ME! */
	@Autowired
	protected MailService mailService;

	/**
	 * Creates a new ORDER_STATES object.
	 */
	public SalesOrderService() {
	}

	/**
	 * state machine which validates if the given sales order can move into the required state.
	 * 
	 * also updates the order at the same time and adds shop notes about this change.
	 * 
	 * staffNumber could also be the payment details source, ie a payment system component like paypal IPN.
	 */
	private void stateTransition(OrderState requiredState, SalesOrder so, StateUpdater updater){
		if(requiredState.equals(OrderState.CREATED)){
			//regardless, this is never allowed!
			throw new IllegalStateException(
					"SO " + so.getUid() + " cannot be put in state " + 
					OrderState.CREATED + " because this is not allowed.");
		}else if(requiredState.equals(OrderState.DISCARDED)){
				//regardless, this is also never allowed!
				throw new IllegalStateException(
						"SO " + so.getUid() + " cannot be put in state " + 
						OrderState.DISCARDED + " because this is not allowed.");
		}else if(requiredState.equals(OrderState.PLACED)){
			if(!getOrderState(so).equals(OrderState.CREATED)){
				throw new IllegalStateException(
						"SO " + so.getUid() + " cannot be put in state " + 
						OrderState.PLACED + " because it is not in the created state.");
			}
			
			String msg = "Status has been set to " + requiredState.getName() + " by " + updater + ".";
			NotesFormatter.appendNotes(msg, so);
			so.setStatusOrderPlaced(new Date());
			
		}else if(requiredState.equals(OrderState.AWAITING_PICKING)){
			if(!getOrderState(so).equals(OrderState.PLACED)){
				throw new IllegalStateException(
						"SO " + so.getUid() + " cannot be put in state " + 
						OrderState.AWAITING_PICKING + " because it is not in the placed state.");
			}

			String msg = "Status has been set to " + requiredState.getName() + " by " + updater + ".";
			NotesFormatter.appendNotes(msg, so);
			so.setStatusAwaitingPicking(new Date());
		
		}else if(requiredState.equals(OrderState.STARTED_PICKING)){
			
			if(!getOrderState(so).equals(OrderState.AWAITING_PICKING)){
				throw new IllegalStateException(
						"SO " + so.getUid() + " cannot be put in state " + 
						OrderState.STARTED_PICKING + " because it is not in the awaiting picking state.");
			}
			
			String msg = "Status has been set to " + requiredState.getName() + " by " + updater + ".";
			NotesFormatter.appendNotes(msg, so);
			so.setStatusStartedPicking(new Date());
			
		}else if(requiredState.equals(OrderState.FINISHED_PICKING)){
			if(!getOrderState(so).equals(OrderState.STARTED_PICKING)){
				throw new IllegalStateException(
						"SO " + so.getUid() + " cannot be put in state " + 
						OrderState.FINISHED_PICKING + " because it is not in the started picking state.");
			}

			String msg = "Status has been set to " + requiredState.getName() + " by " + updater + ".";
			NotesFormatter.appendNotes(msg, so);
			so.setStatusFinishedPicking(new Date());
		
		}else if(requiredState.equals(OrderState.READY_FOR_DISPATCH)){
			
			if(!getOrderState(so).equals(OrderState.FINISHED_PICKING)){
				throw new IllegalStateException(
						"SO " + so.getUid() + " cannot be put in state " + 
						OrderState.READY_FOR_DISPATCH + " because it is not in the finished picking state.");
			}
			
			String msg = "Status has been set to " + requiredState.getName() + " by " + updater + ".";
			NotesFormatter.appendNotes(msg, so);
			so.setStatusReadyForDispatch(new Date());
			
		}else if(requiredState.equals(OrderState.DISPATCHED)){
			if(!getOrderState(so).equals(OrderState.READY_FOR_DISPATCH)){
				throw new IllegalStateException(
						"SO " + so.getUid() + " cannot be put in state " + 
						OrderState.DISPATCHED + " because it is not in the ready for dispatch state.");
			}
			
			String msg = "Status has been set to " + requiredState.getName() + " by " + updater + ".";
			NotesFormatter.appendNotes(msg, so);
			so.setStatusDispatched(new Date());
			
		}else if(requiredState.equals(OrderState.DELIVERED)){
			if(!getOrderState(so).equals(OrderState.DISPATCHED)){
				throw new IllegalStateException(
						"SO " + so.getUid() + " cannot be put in state " + 
						OrderState.DELIVERED + " because it is not in the dispatched state.");
			}

			String msg = "Status has been set to " + requiredState.getName() + " by " + updater + ".";
			NotesFormatter.appendNotes(msg, so);
			so.setStatusDelivered(new Date());
		
		}else if(requiredState.equals(OrderState.CANCELLED)){
			
			if(getOrderState(so).equals(OrderState.DISPATCHED)){
				throw new IllegalStateException(
						"SO " + so.getUid() + " cannot be put in state " + 
						OrderState.CANCELLED + " because it is has already been dispatched.");
			}
			if(getOrderState(so).equals(OrderState.DELIVERED)){
				throw new IllegalStateException(
						"SO " + so.getUid() + " cannot be put in state " + 
						OrderState.CANCELLED + " because it is has already been delivered.");
			}
			if(getOrderState(so).equals(OrderState.CANCELLED)){
				throw new IllegalStateException(
						"SO " + so.getUid() + " cannot be put in state " + 
						OrderState.CANCELLED + " because it is has already been cancelled.");
			}
			if(getOrderState(so).equals(OrderState.DISCARDED)){
				throw new IllegalStateException(
						"SO " + so.getUid() + " cannot be put in state " + 
						OrderState.CANCELLED + " because it is has been discarded (it was never placed).");
			}
			if(getOrderState(so).equals(OrderState.CREATED)){
				throw new IllegalStateException(
						"SO " + so.getUid() + " cannot be put in state " + 
						OrderState.CANCELLED + " because it is has only been created (it was never placed).");
			}

			String msg = "Status has been set to " + requiredState.getName() + " by " + updater + ".";
			NotesFormatter.appendNotes(msg, so);
			so.setStatusCancelled(new Date());
		
		}else if(requiredState.equals(OrderState.PAYMENT_PENDING)){
			throw new IllegalStateException(
					"SO " + so.getUid() + " cannot be put in state " + 
					OrderState.PAYMENT_PENDING + " because that makes no sense.");
		}else if(requiredState.equals(OrderState.PAYMENT_FAILED)){
			//workflow allows setting to failed at any time

			String msg = "Payment status has been set to " + requiredState.getName() + " by " + updater + "."
				+ "The customer should be contacted to arrange a different payment method. "
				+ "When payment is successful, please manually update the order.";
			NotesFormatter.appendNotes(msg, so);
			so.setStatusPaymentFailed(new Date());

		}else if(requiredState.equals(OrderState.PAYMENT_SUCCESSFUL)){
			//workflow allows setting to successful at any time

			String msg = "Payment status has been set to " + requiredState.getName() + " by " + updater + ".";
			NotesFormatter.appendNotes(msg, so);
			so.setStatusPaymentSuccessful(new Date());

		}else{
			throw new IllegalArgumentException("unknown state requiring code change: " + requiredState);
		}
	}

	public OrderState getOrderState(SalesOrder so){
		if(so.getStatusDiscarded() != null){
			return OrderState.DISCARDED;
		}else if(so.getStatusCancelled() != null){
			return OrderState.CANCELLED;
		}else if(so.getStatusDelivered() != null){
			return OrderState.DELIVERED;
		}else if(so.getStatusDispatched() != null){
			return OrderState.DISPATCHED;
		}else if(so.getStatusReadyForDispatch() != null){
			return OrderState.READY_FOR_DISPATCH;
		}else if(so.getStatusFinishedPicking() != null){
			return OrderState.FINISHED_PICKING;
		}else if(so.getStatusStartedPicking() != null){
			return OrderState.STARTED_PICKING;
		}else if(so.getStatusAwaitingPicking() != null){
			return OrderState.AWAITING_PICKING;
		}else if(so.getStatusOrderPlaced() != null){
			return OrderState.PLACED;
		}else if(so.getStatusOrderCreated() != null){
			return OrderState.CREATED;
		}else {
			throw new IllegalStateException("unable to determine state of SO " + so.getUid());
		}
	}
	
	public OrderState getPaymentState(SalesOrder so){
		if(so.getStatusPaymentSuccessful() != null){
			if(so.getStatusPaymentFailed() != null){
				//depends which is newer
				if(so.getStatusPaymentFailed().before(so.getStatusPaymentSuccessful())){
					return OrderState.PAYMENT_SUCCESSFUL;
				}else{
					return OrderState.PAYMENT_FAILED;
				}
			}else{
				return OrderState.PAYMENT_SUCCESSFUL;
			}
		}else if(so.getStatusPaymentFailed() != null){
			return OrderState.PAYMENT_FAILED;
		}else {
			return OrderState.PAYMENT_PENDING;
		}
	}
	
	/**
	 * DOCUMENT ME!
	 *
	 * @param sessionFactory DOCUMENT ME!
	 */
	@Autowired
	public void setSessionFactory(SessionFactory sessionFactory) {
		this.hibernateTemplate = new HibernateTemplate(sessionFactory);
	}

	/**
	 * loads an existing sales order, based on its uid.
	 *
	 * @param  uid the order to load
	 *
	 * @return the fully loaded order, or null if not found.
	 *
	 * @throws IllegalStateException DOCUMENT ME!
	 */
	public SalesOrder getSalesOrder(Integer uid) {
		@SuppressWarnings("unchecked")
		List<SalesOrder> sos = this.hibernateTemplate.find("from SalesOrder so where so.uid = ?", uid);

		if (sos.size() == 1) {
			SalesOrder so = sos.get(0);

			return so;
		} else if (sos.isEmpty()) {
			return null;
		} else {

			// erm more than one found with the same uid?!
			throw new IllegalStateException("more than one sales order with uid " + uid + " found!?");
		}
	}

	/**
	 * loads all existing sales orders, which are placed but not cancelled, not delivered and not dispatched,
	 * ie. confirmed and waiting for something to be done.
	 *
	 * @return the relevant orders, fully loaded, orderd by order number ascending.
	 *
	 * @throws ApplicationException
	 */
	@SuppressWarnings("unchecked")
	public List<SalesOrder> getPendingSalesOrders() throws ApplicationException {

		try {

			// its assumed that the order in which states are set is correct.
			// a state machine checks this before saving order states.
			// hence we do not need to check ALL states, just the relevant ones.
			List<SalesOrder> sos = this.hibernateTemplate.find("from SalesOrder so where "
					+ "so.statusOrderPlaced is not null and "
					+ "so.statusDispatched is null and "
					+ "so.statusCancelled is null "
					+ "order by so.uid asc"); // order asc, so that oldest appear at the top

			return sos;
		} catch (Exception e) {
			String msg = "We failed to get pending sales orders";
			log.error(msg, e);
			throw new ApplicationException("SOS001", msg, "An error occurred. Please try again.");
		}
	}

	/**
	 * loads all discarded sales orders.
	 *
	 * @return the relevant orders, fully loaded, orderd by order number descending.
	 *
	 * @throws ApplicationException
	 */
	@SuppressWarnings("unchecked")
	public List<SalesOrder> getDiscardedSalesOrders() throws ApplicationException {
		
		try {
			
			List<SalesOrder> sos = this.hibernateTemplate.find("from SalesOrder so where "
					+ "so.statusDiscarded is not null "
					+ "order by so.uid desc"); // order desc, so that oldest appear at the bottom
			
			return sos;
		} catch (Exception e) {
			String msg = "We failed to get pending sales orders";
			log.error(msg, e);
			throw new ApplicationException("SOS002", msg, "An error occurred. Please try again.");
		}
	}
	
	/**
	 * loads all existing sales orders, which are problematic. 
	 * ie marked as needing attention or payment failed / pending.
	 * cancelled are not displayed.
	 * discarded are not displayed.
	 *
	 * @return the relevant orders, fully loaded, orderd by order number descending.
	 *
	 * @throws ApplicationException
	 */
	@SuppressWarnings("unchecked")
	public List<SalesOrder> getProblemSalesOrders() throws ApplicationException {
		
		try {
			
			// its assumed that the order in which states are set is correct.
			// a state machine checks this before saving order states.
			// hence we do not need to check ALL states, just the relevant ones.
			List<SalesOrder> result = new ArrayList<SalesOrder>();
			List<SalesOrder> sos = this.hibernateTemplate.find("from SalesOrder so where " +
					"(" +
					"	so.statusPaymentFailed is not null " + //may be relevant - we can test that below, since we need to check if failed is after success (if success aint null)
					"" + 
					"   OR (so.statusPaymentFailed is null AND " +
					"     so.statusPaymentSuccessful is null) " + //these two lines == pending
					"" +
					"	OR so.needsAttention = true " + //always show these ones!
					") " +
					"AND " +
					"so.statusCancelled is null AND " +
					"so.statusDiscarded is null " +
					"order by so.uid desc"); // order desc, so that oldest appear at the bottom
			
			// avoid lazy initialization exceptions!
			for (SalesOrder so : sos) {
				OrderState ps = getPaymentState(so);
				if(so.isNeedsAttention() || 
						ps.equals(OrderState.PAYMENT_PENDING) ||
						ps.equals(OrderState.PAYMENT_FAILED)
						){

					result.add(so);
				}
				
			}
			
			return result;
		} catch (Exception e) {
			String msg = "We failed to get pending sales orders";
			log.error(msg, e);
			throw new ApplicationException("SOS003", msg, "An error occurred. Please try again.");
		}
	}
	
	/**
	 * loads all sales orders, matching the given search criteria.
	 *
	 * @param  search DOCUMENT ME!
	 *
	 * @return the relevant orders, fully loaded, orderd by order number descending.
	 *
	 * @throws ApplicationException
	 */
	@SuppressWarnings("unchecked")
	public List<SalesOrder> searchSalesOrders(String search) throws ApplicationException {

// List<SalesOrder> sos =
// (List<SalesOrder>)hibernateTemplate.executeWithNativeSession(new HibernateCallback() {
// public Object doInHibernate(Session session) throws HibernateException,
// SQLException {
// try{
// List<SalesOrder> sos = new ArrayList<SalesOrder>();
// int rowStart = 0;
// Query q = session.createQuery("from SalesOrder so");
// boolean finished = false;
// while(!finished ) {
// ScrollableResults scr = q.scroll();
// scr.first();
// scr.scroll(rowStart);
// int i = 0;
// databaseList.add(scr.get(0));
// if (!scr.next()){
// break;
// }
// }
// return sos;
// }catch(HibernateException he){
// throw hibernateTemplate.convertHibernateAccessException(he);
// }
// }
// });

		try {

			Integer uid = 0;

			try {
				uid = Integer.parseInt(search);
			} catch (NumberFormatException nfe) {
				// oh well, not a number, so skip UID searching
			}

			// do a wildcard search
			search = "%" + search + "%";

			// now search for artist attributes and pass back all relevant artist
			Object[] params = new Object[12];
			int i = 0;
			params[i ++] = uid; // customer uid
			params[i ++] = search; // customer login
			params[i ++] = search; // customer firstname
			params[i ++] = search; // customer lastname
			params[i ++] = search; // customer screenname
			params[i ++] = search; // customer notes

// params[i++] = uid; //staff uid
// params[i++] = search; //staff firstname
// params[i++] = search; //staff lastname
			params[i ++] = search; // email
			params[i ++] = search; // address postcode
			params[i ++] = search; // shopNotes
			params[i ++] = search; // customerNotes
			params[i ++] = search; // txn id
			params[i ++] = uid; // order id

			List<SalesOrder> sos = this.hibernateTemplate.find("from SalesOrder so where "
					+ "so.partyByCustomerUid.uid like ? or "
					+ "so.partyByCustomerUid.login like ? or "
					+ "so.partyByCustomerUid.firstname like ? or "
					+ "so.partyByCustomerUid.lastname like ? or "
					+ "so.partyByCustomerUid.screenname like ? or "
					+ "so.partyByCustomerUid.customerNotes like ? or "
					+ "so.email like ? or "
					+ "so.address.postcode like ? or "
					+ "so.shopNotes like ? or "
					+ "so.customerNotes like ? or "
					+ "so.txnId like ? or "
					+ "so.uid like ? "
					+ "order by so.uid desc", params); // order desc, so that newest appear at the top

			return sos;
		} catch (Exception e) {
			String msg = "We failed to search sales orders";
			log.error(msg, e);
			throw new ApplicationException("SOS004", msg, "An error occurred. Please try again.");
		}
	}

	/**
	 * loads all customers orders.
	 *
	 * @param  user    DOCUMENT ME!
	 *
	 * @return the relevant orders, fully loaded, orderd by order number descending.
	 *
	 * @throws ApplicationException
	 */
	@SuppressWarnings("unchecked")
	public List<SalesOrder> getCustomerSalesOrders(Party user) throws ApplicationException {
		List<SalesOrder> sos = this.hibernateTemplate.find(
				"from SalesOrder so where "
				+ "so.partyByCustomerUid.uid = ? and "
				+ "so.statusOrderPlaced is not null "
				+ "order by so.uid desc", user.getUid()); // order desc, so that newest appear at the top

		return sos;
	}

	/**
	 * saves an order with status "created".
	 *
	 * @param  basket  the basket to save.
	 * @param  address the delivery address
	 * @param  user    optionally, a logged in user.
	 *
	 * @return a newly saved sales order with its own unique order id.
	 *
	 * @throws ApplicationException if anything does wrong.
	 */
	public SalesOrder createOrder(List<BasketItem> basket, Party user) throws ApplicationException {
		return createOrder(basket, user, false);
	}

	/**
	 * DOCUMENT ME!
	 *
	 * @param  basket    DOCUMENT ME!
	 * @param  address   DOCUMENT ME!
	 * @param  user      DOCUMENT ME!
	 * @param  discarded DOCUMENT ME!
	 *
	 * @return DOCUMENT ME!
	 *
	 * @throws ApplicationException DOCUMENT ME!
	 */
	public SalesOrder createOrder(List<BasketItem> basket, Party user, boolean discarded) throws ApplicationException {
		SalesOrder so = new SalesOrder();

		so.setPartyByCustomerUid(user);
		so.setEmail(user.getLogin());

		if (discarded) {
			so.setStatusDiscarded(new Date());
		} else {
			so.setStatusOrderCreated(new Date());
		}

		Integer salesOrderUid = (Integer) this.hibernateTemplate.save(so);
		so.setUid(salesOrderUid);

		// save duplication - count quantity as we go along
		Map<Integer, SalesOrderItem> items = new HashMap<Integer, SalesOrderItem>();

		MathContext mc = new MathContext(12);

		for (BasketItem bi : basket) {
			SalesOrderItem soi = items.get(bi.getDesign().getUid());

			if (soi == null) {
				soi = new SalesOrderItem();
				soi.setSalesOrder(so);
				soi.setDesign(bi.getDesign());
				soi.setBasePrice(bi.getDesign().getBasePrice().round(mc));
				soi.setAdditionalDeliveryCharge(new BigDecimal("0.00").round(mc)); 
				soi.setSpecialOfferDiscount(new BigDecimal("0.00").round(mc));

				// store against its design ID
				items.put(bi.getDesign().getUid(), soi);

				// add to SO
				so.getSalesOrderItems().add(soi);
			}

			// increment quantity
			soi.setQuantity(soi.getQuantity() + 1);
		}

		// now save the SOIs
		for (SalesOrderItem soi : so.getSalesOrderItems()) {
			this.hibernateTemplate.save(soi);
		}

		so.setDeliveryTotal(new BigDecimal("0.00")); 

		return so;
	}

	/**
	 * @param  orderNum the order that should be marked as needing attention
	 * @param  msg      the message to append to the order notes.
	 *
	 * @throws ApplicationException if something goes wrong.
	 */
	public void markOrderAsNeedingAttention(Integer orderNum, String msg) throws ApplicationException {
		markOrderAsNeedingAttention(null, orderNum, msg);
	}
	
	/**
	 * @param staff the user performing this operation.
	 * @param  orderNum the order that should be marked as needing attention
	 * @param  msg      the message to append to the order notes.
	 *
	 * @throws ApplicationException if something goes wrong.
	 */
	public void markOrderAsNeedingAttention(Party staff, Integer orderNum, String msg)
		throws ApplicationException {
		SalesOrder so = (SalesOrder) this.hibernateTemplate.load(SalesOrder.class, orderNum);
		so.setNeedsAttention(true);
		if(staff == null){
			NotesFormatter.appendNotes(msg, so);
		}else{
			NotesFormatter.appendNotes(staff, msg, so);
		}

		if (configService.boolValue(ConfigService.emailOpsOnMarkOrderAsNeedingAttention)) {

			// email to ops
			msg = "Hi Ops,<br><br>"
				+ "Order " + so.getUid() + " has been marked as needing attention. Please look into this.";
			mailService.sendMail(configService.value(ConfigService.emailOps),
				"[Order Requires Attention] Order " + so.getUid(), msg);
		}
	}

	/**
	 * Payment status has changed. This method should ONLY be called from an automated service like PayPal PDT
	 * or PayPal IPN because it assumes that someone needs to be informed in certain cases. If the payment
	 * status is set to completed or failed using the GUI, then call those methods directly.
	 *
	 * @param  details DOCUMENT ME!
	 *
	 * @throws Exception            If something goes wrong.
	 * @throws ApplicationException DOCUMENT ME!
	 */
	public void updatePaymentStatus(PaymentDetails details) throws ApplicationException {
		log.info("Updating payment status for order " + details.getOrderNumber() + "...");

		// try and ensure that the details contain the order number properly.
		try {
			String badOrderUIDMsg = null;

			if (details.getInvoice() != null) {

				try {
					int invoice = Integer.parseInt(details.getInvoice());

					if ((details.getOrderNumber() == null) || (details.getOrderNumber() == 0)) { // its not been set
						details.setOrderNumber(invoice);
					} else {

						// check they are the same -> e.g. PDT sets the orderUID from the session
						if (invoice != details.getOrderNumber().doubleValue()) {
							badOrderUIDMsg = "The invoice number sent from PayPal PDT or PayPal IPN does not "
								+ "match the orderUID which was stored in the session. The orderUID is "
								+ details.getOrderNumber() + " and the invoice number from PayPal is "
								+ invoice + ". The PayPal invoice number will be used to process this "
								+ "payment update because it is less likely to be wrong. The HTTP session "
								+ "could have expired or something. Please investigate. "
								+ "Determine the customers email address from the order details if required!";

							// use invoice, because its less likely to be wrong - the session could have
							// been lost!
							details.setOrderNumber(invoice);
						}
					}
				} catch (Exception e) {
					badOrderUIDMsg = "The invoice given by PayPal PDT or PayPal IPN is not an integer! It is "
						+ details.getInvoice() + ". The order number stored in the session ("
						+ details.getOrderNumber() + " will be used to process this "
						+ "payment update. Please investigate. "
						+ "Determine the customers email address from the order details if required!";
				}
			} else {

				// this is normal with PDT - lets just check that we have the orderUID in the details
				if (! details.getSource().equals("PayPal PDT")) {
					badOrderUIDMsg = "The invoice number was not supplied when it was expected! "
						+ "The order number stored in the session (" + details.getOrderNumber()
						+ " will be used to process this "
						+ "payment update. Please investigate. "
						+ "Determine the customers email address from the order details if required!";
				}

				if (details.getOrderNumber() == null) {
					throw new Exception(
						"It was not possible to update payment details for the order, because it "
						+ "was not possible to determine the order number. The invoice number was not supplied, "
						+ "and the orderNumber was not available. "
						+ "The invoice was " + details.getInvoice() + " and the order number was "
						+ details.getOrderNumber()
						+ "\r\n\r\nThe full technical transaction details are:\r\n" + details.toString());
				}
			}

			// get the existing order and check the status
			SalesOrder so = (SalesOrder) hibernateTemplate.load(SalesOrder.class, details.getOrderNumber());
			
			// if there was a problem with order UIDs, lets mark the order as needing attention
			if (badOrderUIDMsg != null) {
				markOrderAsNeedingAttention(details.getOrderNumber(), badOrderUIDMsg);
			}

			// check that txnId is correct, unless we are forcing a new one
			if (so.getTxnId() != null) {
				
				if (! details.getTxnId().equals(so.getTxnId())) {
					
					// lets mark the order as needing attention
					String msg =
						"The transaction ID in the payment details does not match the transaction ID stored "
						+ "for this order (" + so.getTxnId()
						+ "). Please investigate this with the payment authority."
						+ "\r\n\r\nThe full technical transaction details are:\r\n" + details.toString();
					markOrderAsNeedingAttention(details.getOrderNumber(), msg);
				}
			} else {
				
				// set it into the order since its currently unknown
				so.setTxnId(details.getTxnId());
			}

			boolean receiverEmailNotMatched = false;
			boolean businessEmailNotMatched = false;

			// check that receiverEmail is your Primary PayPal email
			if (details.getReceiverEmail() != null) {

				if (
					! details.getReceiverEmail().equals(
							configService.value(ConfigService.paypal_merchant_email))) {
					receiverEmailNotMatched = true;

					// lets mark the order as needing attention
					String msg = "The receiver email address provided in the payment details does not "
						+ "match the one which this application is configured for. Please investigate "
						+ "why this is the case. Potentially we have received payment details for a "
						+ "different merchant."
						+ "\r\n\r\nThe full technical transaction details are:\r\n" + details.toString();
					markOrderAsNeedingAttention(details.getOrderNumber(), msg);
				}
			} else {

				// if its PayPal PDT, this is fine
				if (! details.getSource().equals("PayPal PDT")) {
					receiverEmailNotMatched = true;
				}
			}

			// check that receiverEmail is your Primary PayPal email
			if (details.getBusiness() != null) {

				if (
					! details.getBusiness().equals(
							configService.value(ConfigService.paypal_merchant_email))) {
					businessEmailNotMatched = true;

					// lets mark the order as needing attention
					String msg = "The business email address provided in the payment details does not "
						+ "match the one which this application is configured for. Please investigate "
						+ "why this is the case. Potentially we have received payment details for a "
						+ "different merchant."
						+ "\r\n\r\nThe full technical transaction details are:\r\n" + details.toString();
					markOrderAsNeedingAttention(details.getOrderNumber(), msg);
				}
			} else {

				// if its PayPal PDT, this is fine
				if (! details.getSource().equals("PayPal PDT")) {
					businessEmailNotMatched = true;
				}
			}

			if (receiverEmailNotMatched && businessEmailNotMatched) { 

				// lets mark the order as needing attention
				String msg =
					"Neither the business email addres nor the receiver email address provided in the "
					+ "payment details could be matched with the pay_pal_merchant_email set up in the "
					+ "configuration. It is possible that this payment was not for your company! Please "
					+ "check that the order value (mcGross in the technical details below) matches the "
					+ "grand total value of the order. If not, then check if the payerEmail "
					+ "matches the customers email address (" + so.getEmail()
					+ ") - although there is no guarantee "
					+ "that these emails will match. It is advised to contact the payment authority "
					+ "to help determine what has gone wrong with this order payment."
					+ "\r\n\r\nThe full technical transaction details are:\r\n" + details.toString();
				markOrderAsNeedingAttention(details.getOrderNumber(), msg);
			}

			// check that paymentAmounts are correct
			double mcGross = Double.parseDouble(details.getMcGross());
			double grandTotal = SalesOrderHelper.getGrandTotal(so).doubleValue();

			if (mcGross != grandTotal) {

				// lets mark the order as needing attention - POSSIBLE hacking going on as paypal redirect
				// is done over javascript
				String msg = "The gross value of the order in the payment details (" + details.getMcGross()
					+ ") "
					+ "does not match the gross amount stored in the database (" + grandTotal + "). This is "
					+ "strange and should be investigated."
					+ "\r\n\r\nThe full technical transaction details are:\r\n" + details.toString();
				markOrderAsNeedingAttention(details.getOrderNumber(), msg);
			}

			BigDecimal shipping = new BigDecimal(details.getMcShipping());

			if (! shipping.equals(so.getDeliveryTotal())) {

				// lets mark the order as needing attention - POSSIBLE hacking going on as paypal redirect
				// is done over javascript
				String msg = "The shipping costs of the order in the payment details ("
					+ details.getMcShipping() + ") "
					+ "does not match the shipping cost stored in the database (" + so.getDeliveryTotal()
					+ "). This is "
					+ "strange and should be investigated."
					+ "\r\n\r\nThe full technical transaction details are:\r\n" + details.toString();
				markOrderAsNeedingAttention(details.getOrderNumber(), msg);
			}

			// check that paymentCurrencies are correct - POSSIBLE hacking going on as paypal redirect is
			// done over javascript
			if ((details.getMcCurrency() != null) && (so.getPpMcCurrency() != null)
					&& ! details.getMcCurrency().trim().equalsIgnoreCase(so.getPpMcCurrency().trim())) {

				// lets mark the order as needing attention
				String msg = "The payment currency in the payment details (" + details.getMcCurrency() + ") "
					+ "does not match the payment currency stored in the database (" + so.getPpMcCurrency()
					+ "). "
					+ "This is strange and should be investigated."
					+ "\r\n\r\nThe full technical transaction details are:\r\n" + details.toString();
				markOrderAsNeedingAttention(details.getOrderNumber(), msg);
			}

			// if not already there, put currency into DB
			if (so.getPpMcCurrency() == null) {
				so.setPpMcCurrency(so.getPpMcCurrency());
			}

			// put all the new details into the order
			NotesFormatter.appendNotes(details.toString(), so);

			// if the order has not been 'placed', lets do that now
			if (so.getStatusOrderPlaced() == null) {

				// we need to set this now. this method also puts payment details into the order.
				orderPlaced(details, so);
			}

			// process payment by switching on the status
			String status = details.getPaymentStatus().toLowerCase().trim();

			if (status.equals("canceled-reversal")) {

				// lets mark the order as needing attention
				String msg =
					"A reversal has been canceled. For example, you won a dispute with the customer, "
					+ "and the funds for the transaction that was reversed have been returned to you. "
					+ "It is advised that you check the funds to ensure that they have been returned. Check the "
					+ "status of the order to determine if the goods were shipped and if they were, investigate "
					+ "whether they are being returned. See the order history below as there may be details relating "
					+ "to this.\r\n\r\nThe full technical transaction details are:\r\n" + details.toString();
				markOrderAsNeedingAttention(details.getOrderNumber(), msg);
			} else if (status.equals("completed")) {
				paymentSuccessful(so, StateUpdater.createPaymentSystem(details.getSource()));
			} else if (status.equals("denied")) {

				// lets mark the order as needing attention
				String msg = "You denied the payment. This happens only if the payment was "
					+ "previously pending because of possible reasons described for the PendingReason element. "
					+ "Please ensure that the customer has been informed of this. "
					+ "If required, also cancel this order. "
					+ "See the order history below as there may be details relating "
					+ "to this.\r\n\r\nThe full technical transaction details are:\r\n" + details.toString();
				markOrderAsNeedingAttention(details.getOrderNumber(), msg);
			} else if (status.equals("expired")) {

				// lets mark the order as needing attention
				String msg = "This authorization has expired and cannot be captured. "
					+ "It is advised that you check the funds to see if they have indeed been paid. "
					+ "It is likely that a new method of payment will need to be arranged. "
					+ "See the order history below as there may be details relating "
					+ "to this.\r\n\r\nThe full technical transaction details are:\r\n" + details.toString();
				markOrderAsNeedingAttention(details.getOrderNumber(), msg);
			} else if (status.equals("failed")) {
				paymentFailed(so, StateUpdater.createPaymentSystem(details.getSource()));

				// now, because this method is called from an automated system, lets flag the order
				// as needing attention! See JavaDocs above!!
				String msg = "Payment has failed for this order. "
					+ "The customer should be contacted to arrange a different method of payment. See the order "
					+ "history for details.";
				markOrderAsNeedingAttention(so.getUid(), msg);
			} else if (status.equals("in-progress")) {

				// lets just add the details to the order history
				String msg = "The transaction is in process of authorization and capture."
					+ "\r\n\r\nThe full technical transaction details are:\r\n" + details.toString();
				NotesFormatter.appendNotes(msg, so);
			} else if (status.equals("partially-refunded")) {

				// lets mark the order as needing attention
				String msg = "The transaction has been partially refunded. "
					+ "It is advised that you check below to see the order history to understand "
					+ "why the transaction has been partially refunded. Contact the customer if "
					+ "required."
					+ "\r\n\r\nThe full technical transaction details are:\r\n" + details.toString();
				markOrderAsNeedingAttention(details.getOrderNumber(), msg);
			} else if (status.equals("pending")) {

				// lets mark the order as needing attention
				String msg = "The payment is pending. It is likely that this "
					+ "order will need manual intervention. The pending reason is:\r\n\r\n"
					+ details.getPendingReason()
					+ "\r\n\r\nSee the order history below as there may be details relating "
					+ "to this.\r\n\r\nThe full technical transaction details are:\r\n" + details.toString();
				markOrderAsNeedingAttention(details.getOrderNumber(), msg);
			} else if (status.equals("processed")) {

				// this is when the merchant is sent money and they choose to accept the payment.
				// hence its the same as a successful payment
				paymentSuccessful(so, StateUpdater.createPaymentSystem(details.getSource()));
			} else if (status.equals("refunded")) {

				// lets mark the order as needing attention
				String msg =
					"You refunded the payment. Check in the history below to understand why this is. "
					+ "If the goods need to be returned, check that this has occurred before setting the order "
					+ "status to cancelled. Alternatively, it is possible that someone in the company has "
					+ "authorized that the customer retain the goods. "
					+ "See the order history below as there may be details relating "
					+ "to this.\r\n\r\nThe full technical transaction details are:\r\n" + details.toString();
				markOrderAsNeedingAttention(details.getOrderNumber(), msg);
			} else if (status.equals("reversed")) {

				// lets mark the order as needing attention
				String msg = "A payment was reversed due to a chargeback or other type of "
					+ "reversal. The funds have been removed from your account balance and "
					+ "returned to the buyer. The reason for the reversal is specified in the"
					+ "ReasonCode element in the technical transaction details below. "
					+ "There should be dispute details in the history below to help understand this "
					+ "payment reversal. If there are none, it is advised to contact the payment authority. "
					+ "If goods are to be returned for this order, do not set it to cancelled until they arrive. "
					+ "See the order history below as there may be details relating "
					+ "to this.\r\n\r\nThe full technical transaction details are:\r\n" + details.toString();
				markOrderAsNeedingAttention(details.getOrderNumber(), msg);
			} else if (status.equals("voided")) {
				String msg =
					"This authorization has been voided. This is likely to mean that payment has failed. "
					+ "Check the accounts to determine if payment has actually been received. If it has not, then "
					+ "a different method of payment will need to be arranged for this order. If this is the case, "
					+ "set the order status to 'Placed'. "
					+ "See the order history below as there may be details relating "
					+ "to this.\r\n\r\nThe full technical transaction details are:\r\n" + details.toString();
				markOrderAsNeedingAttention(details.getOrderNumber(), msg);
			} else {
				throw new Exception("Unexpected payment status " + status
					+ ", meaning that there is an error in the framework. Please raise this with technical support immediately!");
			}

		} catch (Exception e) {
			log.error("Failed to update payment status", e);
			throw new ApplicationException("SOS006",
				"We failed to update the payment status for order " + details.getOrderNumber(),
				"An error occurred. Please try again.");
		}
	}

	/**
	 * When the order details have been collected, call this method. E.g. from manual entry of details, like
	 * for an invoice order, or from something like PayPal IPN.
	 *
	 * @param  details orderUID
	 * @param  so      DOCUMENT ME!
	 *
	 * @throws ApplicationException DOCUMENT ME!
	 */
	private void orderPlaced(PaymentDetails details, SalesOrder so) throws ApplicationException {
		
		//check we can update this
		stateTransition(OrderState.PLACED, so, null);
		
		so.setStatusOrderPlaced(new Date());

		String cssStandardText = "font-family: Tahoma, Verdana, Geneva, Arial, Helvetica, sans-serif !important; font-size: 12; "; 
		String cssTableHeader = "background-color: #dddddd; font-family: Tahoma, Verdana, Geneva, Arial, Helvetica, sans-serif !important; font-size: 12; font-weight: bold; "; 
		String cssTableBodyEven = "font-family: Tahoma, Verdana, Geneva, Arial, Helvetica, sans-serif !important; font-size: 12; background-color: #EEEEEE; "; 
		String cssTableBodyOdd = "font-family: Tahoma, Verdana, Geneva, Arial, Helvetica, sans-serif !important; font-size: 12; background-color: #FFFFFF; "; 
		String cssTableBodyTotal = "font-family: Tahoma, Verdana, Geneva, Arial, Helvetica, sans-serif !important; font-size: 12; background-color: #CCCCCC; "; 
		// send mail to customer and copy it to backup account
		String msg =
			"<p style='" + cssStandardText  + "'>"
			+ "Hi " + so.getPartyByCustomerUid().getFirstname() + ",<br><br>"
			+ "Thank you for placing your order with us.<br><br>"
			+ "Order " + so.getUid() + " will be "
			+ "picked and dispatched shortly. We will send you another email when the order is "
			+ "shipped to you.<br><br>"
			+ "The order details were:<br><br>"
			+ "<table border='1' cellpadding='0' cellspacing='0'>"
			+ "<tr >"
			+ "  <th align='center' style='" + cssTableHeader + "'>ID</th>"
			+ "  <th align='center' style='" + cssTableHeader + "'>Name</th>"
			+ "  <th align='center' style='" + cssTableHeader + "'>Qty</th>"
			+ "</tr>";

		int i = 0;

		for (SalesOrderItem soi : so.getSalesOrderItems()) {
			String style = (((i ++ % 2) == 0) ? cssTableBodyOdd : cssTableBodyEven);
			msg += "<tr style='" + style + "'>"
				+ "  <td style='" + style + "'>" + soi.getDesign().getUid() + "</td>"
				+ "  <td style='" + style + "'>" + soi.getDesign().getName() + "</td>"
				+ "  <td style='" + style + "'>" + soi.getQuantity() + "</td>"
				+ "</tr>";
		}

		msg += 
				"<tr style='" + cssTableBodyTotal + "'>" +
				"  <td colspan='3' align='right' style='" + cssTableBodyTotal + "'>Total " +
					ServerFormats.NF_0_00().format(SalesOrderHelper.getGrandTotal(so)) + " GBP</td>" +
				"</tr>"; 
		String url = configService.value(ConfigService.httpUrl);
		msg += "</table><br><br>"
			+ "To view your order, please login at <a href='" + url + "/secure/index.jsp?screen=11&uid=" + so.getUid() + "'>" + url + "</a>.<br><br>"
			+ "more text to be specified by the business"
			+ "</p>"; 
		mailService.sendMail(so.getEmail(), "Order " + so.getUid(), msg);

		if (configService.boolValue(ConfigService.emailOpsOnOrderPlaced)) {

			// ...and merchant
			msg = "Hi Ops,<br><br>"
				+ "Order " + so.getUid() + " has been placed.";
			mailService.sendMail(configService.value(ConfigService.emailOps),
				"[Order Placed] Order " + so.getUid(), msg);
		}
	}

	/**
	 * If either IPN or the EPOS owner or some other gateway informs us of a dispute, call this method to add
	 * the details to the orders notes, and mark this order as needing manual attention.
	 *
	 * @param  details DOCUMENT ME!
	 *
	 * @throws ApplicationException If something goes wrong.
	 */
	public void handleDispute(PaymentDetails details) throws ApplicationException {

		try {
			String msg = "A dispute has been raised over this order.\r\n"
				+ "Based on the details below, txnType should be 'new_case', 'txnId' is your (the merchant) "
				+ "original transaction identification number for the payment from the buyer, against which "
				+ "the case was registered. 'caseId' is the case identification number, of format "
				+ "PP-nnn-nnn-nnn where n is any numeric character. 'caseType' can be either complaint or "
				+ "chargeback - if its 'complaint', then a buyer has logged a complaint through the PayPal "
				+ "Resolution Center - if its 'chargeback', a buyer has filed a chargeback with their credit "
				+ "card company, which has notified PayPal of the reason for the chargeback. 'case_creation_date' "
				+ "is transaction specific, but is the date and time case was registered in format: "
				+ "HH:mm:ss mon DD, YYYY TZ where TZ is a three-character time zone.\r\n"
				+ "\r\n"
				+ "'reason_code' "
				+ "depends on the value of 'case_type' and is the reason for the case. If the 'case_type' "
				+ "is 'complaint' then this can be 'non_receipt' meaning that the buyer claims that they did "
				+ "not receive "
				+ "goods or service or it can be 'not_as_described' means that the buyer claims that the goods or "
				+ "service received differ from the merchants description of the goods or service.\r\n"
				+ "\r\n"
				+ "If the 'case_type' is 'chargeback', then 'reason_code' can have values 'unauthorized' meaning "
				+ "that the buyer claims that he did not authorize transaction, or it can be 'non_receipt' "
				+ "meaning that the buyer claims that he did not receive goods or service. It can also be "
				+ "'duplicate' meaning that the buyer claims that a possible duplicate payment was made to the "
				+ "merchant. It can be 'merchandise' where the buyer claims that the received merchandise is "
				+ "unsatisfactory, defective, or damaged. It can be 'special' meaning there is some other "
				+ "reason for the dispute. Usually, 'special' indicates a credit card processing error for which "
				+ "the merchant is not responsible and for which no debit to the merchant will result. "
				+ "PayPal must review the documentation from the credit card company to determine the nature of "
				+ "the dispute and possibly contact the merchant to resolve it."
				+ "\r\n\r\nThe full technical transaction details are:\r\n" + details.toString();
			markOrderAsNeedingAttention(details.getOrderNumber(), msg);
		} catch (Exception e) {
			log.error("Failed to handle dispute", e);
			throw new ApplicationException("SOS007",
				"Failed to handle dispute for order " + details.getOrderNumber(),
				"An error occurred. Please try again.");
		}
	}

	/**
	 * Sets status of payment to successful, adds notes, then for any gift certificates which were purchased,
	 * they are also updated as being paid for.
	 *
	 * @param  so         the order where the payment was successful
	 * @param  orderTotal DOCUMENT ME!
	 * @param  staff 		the user, if it was a user who marked the payment as failed. null if it was an
	 *                    automated system.
	 *
	 * @throws ApplicationException DOCUMENT ME!
	 */
	public void paymentSuccessful(SalesOrder so, StateUpdater updater)
		throws ApplicationException {

		//check we can update this
		stateTransition(OrderState.PAYMENT_SUCCESSFUL, so, updater);

		if (configService.boolValue(ConfigService.emailOpsOnSuccessfulPayment)) {

			// email to ops - picking/dispatch can occur
			String msg = "Hi Ops,<br><br>"
				+ "Order " + so.getUid() + " has successfully been paid for.";
			mailService.sendMail(configService.value(ConfigService.emailOps),
				"[Successful Payment] Order " + so.getUid(), msg);
		}

		
		// update status...
		if(getOrderState(so).equals(OrderState.PLACED)){
			stateTransition(OrderState.AWAITING_PICKING, so, updater);
		}else{
			//no worries - its already at least awaiting picking
			//ASSUMPTION: we would never ever get here, if the order was only in 
			//the created state, since the UI doesnt show such orders. 
			//see #getPendingSalesOrders()
		}
	}

	/**
	 * marks the order with payment status failed.
	 *
	 * @param  details details of the payment.
	 * @param  so      the sales order.
	 * @param  user    the user, if it was a user who marked the payment as failed. null if it was an
	 *                 automated system.
	 *
	 * @throws ApplicationException DOCUMENT ME!
	 */
	public void paymentFailed(SalesOrder so, StateUpdater updater)
		throws ApplicationException {

		//check we can update this
		stateTransition(OrderState.PAYMENT_FAILED, so, updater);

		if (configService.boolValue(ConfigService.emailOpsOnFailedPayment)) {

			// email to ops - picking/dispatch can occur
			String msg = "Hi Ops,<br><br>"
				+ "Payment for Order " + so.getUid()
				+ " has failed. Please see the order details for more information.";
			mailService.sendMail(configService.value(ConfigService.emailOps),
				"[Failed Payment] Order " + so.getUid(), msg);
		}
	}

	/**
	 * updates the status of this order. PAYMENT STATII ARE NOT ACCEPTED!
	 * shop notes are added about this change.
	 * 
	 * @param staff who is doing this?
	 * @param requiredState which state the order should have afterwards
	 * @param soUid the order to update.
	 * @param notes optional notes to be added to the order. eg if its been cancelled.
	 * @throws ApplicationException if anything goes wrong.
	 */
	public void updateOrderStatus(Party staff, OrderState requiredState, int soUid, String notes) throws ApplicationException {
		log.info("Updating order status for order " + soUid + "...");
		try{
			if(requiredState.equals(OrderState.PAYMENT_FAILED) || 
					requiredState.equals(OrderState.PAYMENT_PENDING) || 
					requiredState.equals(OrderState.PAYMENT_SUCCESSFUL)){
				throw new IllegalArgumentException("Code change required as payment status is not acceptable here.");
			}
			
			SalesOrder so = getSalesOrder(soUid);
			
			stateTransition(requiredState, so, StateUpdater.createStaff(staff));
			
			if(notes != null){
				NotesFormatter.appendNotes(staff, notes, so);
			}
			
		} catch (IllegalStateException e) {
			String msg = "Failed to update order state for order " + soUid + ". " + e.getMessage();
			log.error(msg , e);
			throw new ApplicationException("SOS008",
				msg,
				"Order is in a state which cannot be updated to " + requiredState + ".<br>" +
					"Please try reloading the order first, since someone else may have already changed the order.");
		} catch (Exception e) {
			String msg = "Failed to update order state for order " + soUid + ". " + e.getMessage();
			log.error(msg , e);
			throw new ApplicationException("SOS009",
				msg,
				"An error occurred. Please try again.");
		}
	}

	/**
	 * updates the status of this order. Only PAYMENT STATII ARE ALLOWED!
	 * 
	 * @param staff who is doing this?
	 * @param requiredState which state the order should have afterwards
	 * @param soUid the order to update.
	 * @throws ApplicationException if anything goes wrong.
	 */
	public void updatePaymentStatus(Party staff, OrderState requiredState, int soUid, String notes, String txnId) throws ApplicationException {
		log.info("Updating payment status for order " + soUid + "...");
		try{
			if(!requiredState.equals(OrderState.PAYMENT_FAILED) && 
					!requiredState.equals(OrderState.PAYMENT_PENDING) && 
					!requiredState.equals(OrderState.PAYMENT_SUCCESSFUL)){
				throw new IllegalArgumentException("Code change required as order status is not acceptable here.");
			}

			SalesOrder so = getSalesOrder(soUid);

			String set = "set";
			if(so.getTxnId() != null){
				set = "reset";
			}
			String msg =
				"The transaction ID has been " + set + " to " + txnId + ".";
			if(txnId == null){
				msg = "The transaction ID has been cleared.";
			}
			if(so.getTxnId() != null){
				msg += " It was previously " + so.getTxnId() + ".\r\n\r\n";
			}
			msg += notes;
			NotesFormatter.appendNotes(staff, msg, so);

			so.setTxnId(txnId);
			
			if(requiredState.equals(OrderState.PAYMENT_FAILED)){
				paymentFailed(so, StateUpdater.createStaff(staff));
			}else if(requiredState.equals(OrderState.PAYMENT_SUCCESSFUL)){
				paymentSuccessful(so, StateUpdater.createStaff(staff));
			}else{
				throw new IllegalArgumentException("code change required: unexpected required payment state: " + requiredState);
			}

			
		} catch (IllegalStateException e) {
			String msg = "Failed to update payment state for order " + soUid + ". " + e.getMessage();
			log.error(msg , e);
			throw new ApplicationException("SOS010",
					msg,
					"Order is in a state which cannot be updated to " + requiredState + ".<br>" +
			"Please try reloading the order first, since someone else may have already changed the order.");
		} catch (Exception e) {
			String msg = "Failed to update payment state for order " + soUid + ". " + e.getMessage();
			log.error(msg , e);
			throw new ApplicationException("SOS011",
					msg,
			"An error occurred. Please try again.");
		}
	}

	/**
	 * adds the notes with a nicely formatted header.
	 * 
	 * @param staff who is doing this?
	 * @param soUid the order to update.
	 * @param notes the note sto add
	 * @throws ApplicationException if anything goes wrong.
	 */
	public void addShopNotes(Party staff, int soUid, String notes) throws ApplicationException {
		log.info("Updating notes for order " + soUid + "...");
		try{
			SalesOrder so = getSalesOrder(soUid);

			NotesFormatter.appendNotes(staff, notes, so);

		} catch (Exception e) {
			String msg = "Failed to add notes for order " + soUid + ". " + e.getMessage();
			log.error(msg , e);
			throw new ApplicationException("SOS012",
					msg,
					"An error occurred. Please try again.");
		}
	}

	/**
	 * marks the order as no longer needing attention.
	 * 
	 * @param staff who is doing this?
	 * @param soUid the order to update.
	 * @throws ApplicationException if anything goes wrong.
	 */
	public void markOrderAsNotNeedingAttention(Party staff, int soUid, String notes) throws ApplicationException {
		log.info("Marking order " + soUid + " as NOT needing attention...");
		try{
			SalesOrder so = getSalesOrder(soUid);
			
			String msg = "This order has been marked as NOT needing attention:\r\n\r\n" + notes;
			NotesFormatter.appendNotes(staff, msg, so);
			
			so.setNeedsAttention(false);

		} catch (Exception e) {
			String msg = "Failed to mark order as NOT needing attention for order " + soUid + ". " + e.getMessage();
			log.error(msg , e);
			throw new ApplicationException("SOS014",
					msg,
			"An error occurred. Please try again.");
		}
	}
	
	/**
	 * a person or system who is updating the status of an order.
	 */
	public static class StateUpdater {
		private enum Type {
			PaymentSystem,
			Staff
		}
		
		private Type type;
		private String reference;

		private StateUpdater(String reference, Type type){
			this.reference = reference;
			this.type = type;
		}
		
		static StateUpdater createStaff(Party staff){
			return new StateUpdater(NotesFormatter.getStaffDescription(staff), Type.Staff);
		}
		
		static StateUpdater createPaymentSystem(String reference){
			return new StateUpdater(reference, Type.PaymentSystem);
		}
		
		public String toString(){
			if(type.equals(Type.PaymentSystem)){
				return reference;
			}else if(type.equals(Type.Staff)){
				return "Staff " + reference;
			}else{
				throw new IllegalArgumentException("code change required for unknown status updater type " + type);
			}
		}
	}
}
