/*  
 * Copyright (c) 2009 Ant Kutschera, maxant
 * 
 * This file is part of the maxant GWT Demo.
 * 
 * maxant GWT Demo is free software: you can redistribute it and/or modify
 * it under the terms of the Lesser GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * maxant GWT Demo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Lesser GNU General Public License for more details.
 * You should have received a copy of the Lesser GNU General Public License
 * along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */

package uk.co.maxant.gwtdemo.server.util;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.math.BigDecimal;
import java.util.Date;

/**
 * hibernate objects, whilst serializable, cannot be transferred to a GWT UI because
 * lazily loaded kids are proxies which cannot be deserialized on the client.
 * as a result, the persistent objects need to be mapped to transfer objects.
 * 
 * the next problem, is that a direct field mapping does not work because the lazily
 * loaded child objects which are proxies have fields with different names! so
 * the mapping occurs over accessor methods. all declared methods in the target and 
 * source are compared and any getters/setters with the same names are used to do the
 * mapping, but only of Strings, Integers, Longs, Floats, Doubles and BigDecimals.
 * 
 * Any other child objects needs to be mapped by hand.
 */
public class GenericMapper {

	private GenericMapper(){}

	/**
	 * Maps from source to a new target instance. the new instance is returned.
	 * <b>the new instance is NOT created using a constructor!</b>
	 * 
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException 
	 */
	@SuppressWarnings("unchecked")
	public static Object map(Class targetClass, Object source) throws InstantiationException, IllegalAccessException, InvocationTargetException {
		
		Object target = targetClass.newInstance();
		map(source, target);
		return target;
	}

	/**
	 * Maps from source to target, using setters and getters.
	 * 
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException 
	 */
	public static void map(Object source, Object target) throws InstantiationException, IllegalAccessException, InvocationTargetException {
		
		Method[] fromMethods = source.getClass().getDeclaredMethods();
		for(Method mFrom : fromMethods){
			if(mFrom.getName().startsWith("get")){
				String name = "s" + mFrom.getName().substring(1);
				Method[] toMethods = target.getClass().getDeclaredMethods();
				for(Method mTo : toMethods){
					if(mTo.getName().equals(name)){
						Object obj = mFrom.invoke(source, (Object[])null);
						boolean doit = false;
						if(obj instanceof String){
							doit = true;
						}else if(obj instanceof Integer){
							doit = true;
						}else if(obj instanceof Long){
							doit = true;
						}else if(obj instanceof Double){
							doit = true;
						}else if(obj instanceof Float){
							doit = true;
						}else if(obj instanceof BigDecimal){
							doit = true;
						}else if(obj instanceof Date){
							doit = true;
						}
						if(doit){
							try{
								mTo.invoke(target, obj);
							}catch(IllegalArgumentException e){
								//oh well, attribute aint mapped!
							}
						}
					}
				}
			}
		}

		//the following old code attempted to do direct field mapping, which only worked for
		//non-lazily loaded hibernate POs.
				
		//Field[] fromFields = source.getClass().getDeclaredFields();
		//		
		//for(Field fFrom : fromFields){
		//	try {
		//		Field fTo = target.getClass().getDeclaredField(fFrom.getName());
		//		if(fTo != null){
		//			fFrom.setAccessible(true); //in case its private
		//			fTo.setAccessible(true); //in case its private
		//			fTo.set(target, fFrom.get(source));
		//		}
		//	} catch (NoSuchFieldException e) {
		//		//oh well, skip it since it cannot be mapped to a field in the target
		//	}
		//}
	}
}
