package uk.co.maxant.spring.security;

import java.lang.reflect.Method;

import org.aspectj.lang.ProceedingJoinPoint;
import org.aspectj.lang.reflect.MethodSignature;
import org.springframework.core.Ordered;

/**
 * aspect which should be configured to be inserted before all service calls.
 * it checks the method about to be called for the {@link RolesAllowed} annotation.
 * if the logged in user has one of the allowed roles, the method is called. otherwise
 * a {@link SecurityException} is called.
 */
public class SecurityAspect implements Ordered {

	private int order;

	/**
	 * allows us to control the ordering of advice
	 */
	public int getOrder() {
		return this.order;
	}

	public void setOrder(int order) {
		this.order = order;
	}

	/**
	 * this method is the around advice. it is called prior to the service method being called.
	 * it checks if the user is authenticated and authorised to carry out the service call.
	 */
	@SuppressWarnings("unchecked")
	public Object checkSecurity(ProceedingJoinPoint call) throws Throwable {

		String[] requiredRoles = null;
		MethodSignature ms = (MethodSignature)call.getSignature();
		Method m = ms.getMethod();
		
		//lets check it for the RolesAllowed annotation
		RolesAllowed roles = m.getAnnotation(RolesAllowed.class);
		if(roles != null){
			requiredRoles = roles.value();
			for(String role : requiredRoles){
				if(SecurityHelper.isInRole(role)){
					//ok, no problem, lets continue with the call
					requiredRoles = null; //since we can now carry out the call
					break;
				}
			}
		}

		if(requiredRoles == null){
			//no problems encountered, simply call the method
			Object returnValue = call.proceed();
			return returnValue;
		}else{
			
			//make a pretty string
			String rrs = "[";
			for(String s : requiredRoles){
				rrs += s + ", ";
			}
			rrs = rrs.substring(0, rrs.length() - 2) + "]";
			
			throw new SecurityException("To call [" + call.getSignature().toLongString() + "] the user must be authenticated and authorised in these roles: " + rrs);
		}
	}
}