package uk.co.maxant.spring.security;

import javax.servlet.http.HttpServletRequest;

/**
 * stores request objects mapped to threads, so that the request can be found 
 * anywhere in the callstack. based on thread locals.
 * 
 * call {@link #finished()} when you no longer need to query roles, to help keep
 * memory usage to a minimum, otherwise this class keeps references to requests which
 * have references to sessions which could be quite large.
 */
class SecurityHelper {

	private static ThreadLocal<HttpServletRequest> tls = new ThreadLocal<HttpServletRequest>();
	
	private SecurityHelper(){}

	/**
	 * to be called by the filter to check if the user has the requested role.
	 * @param role the required role to check
	 * @return true, if the user has that role. false if the user does not, or
	 * 					is not logged in, or if the request could not be found
	 * 					because the {@link #set(HttpServletRequest)} method has
	 * 					not been called.
	 */
	static boolean isInRole(String role){
		HttpServletRequest req = tls.get();
		if(req == null){
			return false;
		}else{
			return req.isUserInRole(role);
		}
	}

	/**
	 * to be called by the filter, to set the request into the security helper for the current thread.
	 * @param request the current request object related to this thread.
	 */
	static void set(HttpServletRequest request){
		tls.set(request);
	}

	/**
	 * to be called by the aspect, once all role queries have been finished.
	 */
	static void finished(){
		tls.remove();
	}
	
}
